"""
Module for loading L3 housekeeping data using a database that
indexes an archive of HK files.

"""
from dataclasses import dataclass, field
import yaml
import os
import logging
import time
import numpy as np
import pathlib
import fnmatch
import hashlib

import sqlalchemy as db
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker, relationship
from typing import Union, List, Optional, Dict, Any
import so3g
from spt3g import core as spt3g_core
from tqdm.auto import tqdm

Base = declarative_base()
Session = sessionmaker()

log = logging.getLogger(__name__)

@dataclass
class HkConfig:
    """
    Configuration object for indexing and loading from an HK archive.

    If instantiating from a nested dictionary, the ``from_dict`` class method
    can be used to convert fields to their proper data types.

    Args
    ------------
    hk_root: str
        Root directory for the HK archive
    db_file: Optional[str]
        Path to the hk index database if the database is an sqlite file. Either
        this or db_url must be set
    db_url: Optional[Union[str, db.URL]]
        URL used for db engine. Either this or db_file must be set
    echo_db: bool
        Whether database operations should be echoed
    file_idx_lookback_time: Optional[float]
        Time [sec] to look back when scanning for new files to index
    show_index_pb: bool
        If true, shows progress bar when indexing
    aliases: Dict[str, str]
        Aliases for hk fields. In this dict, the key is the alias name, and the
        value is the field descriptor, in the format of ``agent.feed.field``.
        For example::

            {
                'fp_temp': 'cryo-ls372-lsa21yc.temperatures.Channel_02_T',
            }

        These aliases are only used on load, and do not affect how data is stored in the hkdb.
        Aliases should be valid python identifiers, since they will be set as attributes in
        the HkResult object.
    """
    hk_root: str
    db_file: Optional[str] = None
    db_url: Optional[Union[str, db.URL]] = None
    echo_db: bool = False
    file_idx_lookback_time: Optional[float] = None
    show_index_pb: bool = True
    aliases: Dict[str, str] = field(default_factory=dict)

    def __post_init__(self):
        if self.db_file is None and self.db_url is None:
            raise ValueError("Either db_file or db_url must be set")
        if self.db_file is not None and self.db_url is not None:
            raise ValueError("Only one of db_file or db_url must be set")
        if self.db_file is not None:
            self.db_url = f"sqlite:///{self.db_file}"

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> "HkConfig":
        """
        Generates an HkConfig object from a dictionary whose keys are fields of
        the HkConfig dataclass.

        If the ``db_url`` is specified, it can be set as a string, a
        dictionary, or a sqlalchemy URL object. If it is of type dict, it
        will be converted to a URL object by passing it through to the keyword
        arguments of sqlalchemy's URL.create function. Environment variables
        will be expanded in both the string and dict representations.
        """
        _db_url = data.get('db_url')
        if isinstance(_db_url, dict):
            url_dict = data['db_url']
            for k, v in url_dict.items():
                url_dict[k] = os.path.expandvars(v)
            data['db_url'] = db.URL.create(**url_dict)
        elif isinstance(_db_url, str):
            data['db_url'] = os.path.expandvars(data['db_url'])
        return cls(**data)

    @classmethod
    def from_yaml(cls, path):
        with open(path, 'r') as f:
            return cls.from_dict(yaml.safe_load(f))


class HkFile(Base):
    """
    Database entry for a hk file.

    Args
    ------
    path: str
        Path to the hk file.
    start_time: float
        Starting ctime of the file
    end_time: float
        Ending ctime of the file
    size: int
        Size of file in bytes
    mod_time: float
        Modification time of file
    index_status: float
        Index status of file. Can be "unindexed", "indexed", or "failed"
    """
    __tablename__ = 'hk_files'
    id = db.Column(db.Integer, primary_key=True)
    path = db.Column(db.String, nullable=False, unique=True)
    start_time = db.Column(db.Float)
    end_time = db.Column(db.Float)
    size = db.Column(db.Float)
    mod_time = db.Column(db.Float)
    index_status = db.Column(db.String)


class HkFrame(Base):
    """
    Database entry for an hk frame.

    Args
    --------
    file_id: int
        ID of the file containing the frame
    agent: str
        instance-id of the OCS agent for the frame
    feed: str
        Name of the OCS feed for the frame
    fields_hash: str
        Hash to identify the combination of fields included in this frame.
    byte_offset: int
        Offset of the frame in the G3 file in bytes
    start_time: float
        Starting ctime of the frame
    end_time: float
        Ending ctime of the frame
    """
    __tablename__ = 'hk_frames'
    id = db.Column(db.Integer, primary_key=True)
    file_id = db.Column(db.Integer, db.ForeignKey('hk_files.id'),
                        index=True)  # index required on postgres, or
                                     # else delete is slow
    file = relationship('HkFile')
    agent = db.Column(db.String)
    feed = db.Column(db.String)
    fields_hash = db.Column(db.String)
    byte_offset = db.Column(db.Integer)
    start_time = db.Column(db.Float)
    end_time = db.Column(db.Float)


class HkDb:
    """
    Helper class for createing database sessions

    Args
    ------
    cfg : Union[HKConfig, str]
        Configuration object or path to configuration file
    """
    def __init__(self, cfg: Union[HkConfig, str]):
        if isinstance(cfg, str):
            cfg = HkConfig.from_yaml(cfg)
        self.cfg = cfg

        self.engine = db.create_engine(cfg.db_url, echo=cfg.echo_db)
        Session.configure(bind=self.engine)
        self.Session = sessionmaker(bind=self.engine)
        Base.metadata.create_all(self.engine)


def update_file_index(hkcfg: HkConfig, session=None, subdirs=None,
                      retry_failed=False):
    """Updates HkFiles database with new files on disk.

    If subdirs is specified, it must be a list of (full path)
    sub-directories of hk_root; those will be scanned and any new
    files added to database.

    Otherwise, all subdirs of hk_root will be scanned, subject to any
    restriction from hkcfg.file_idx_lookback_time.

    """
    if session is None:
        hkdb = HkDb(hkcfg)
        session = hkdb.Session()

    if hkcfg.file_idx_lookback_time is not None and subdirs is None:
        min_ctime = time.time() - hkcfg.file_idx_lookback_time
    else:
        min_ctime = 0

    all_files = []
    if subdirs is None:
        log.debug(f"Getting dirs from {hkcfg.hk_root}.")
        subdirs = [os.path.join(hkcfg.hk_root, subdir)
                   for subdir in os.listdir(hkcfg.hk_root)]
    else:
        _root = pathlib.Path(hkcfg.hk_root)
        assert all([_root in pathlib.Path(sdir).parents for sdir in subdirs])

    log.debug(f"Scanning {len(subdirs)} subdirs (min_ctime={min_ctime}).")
    for sdir in subdirs:
        if min_ctime > 0 and min_ctime > os.path.getmtime(sdir):
            continue
        all_files.extend([
            os.path.join(sdir, f)
            for f in os.listdir(sdir)
            if f.endswith('.g3')
        ])

    existing_files = {'all': []}
    for k in ['failed', 'indexed', 'unindexed']:
        existing_files[k] = [
            os.path.join(hkcfg.hk_root, path)
            for path, in session.query(HkFile.path).filter(HkFile.index_status == k).all()
        ]
        existing_files['all'].extend(existing_files[k])

    if retry_failed:
        retry_files = sorted(list(set(all_files).intersection(existing_files['failed'])))
        log.info(f"Clearing {len(retry_files)} for retry...")
        for f in retry_files:
            relpath = os.path.relpath(f, hkcfg.hk_root)
            session.query(HkFile).filter(HkFile.path == relpath) \
                                 .update({HkFile.index_status: 'unindexed'})

    new_files = sorted(list(set(all_files) - set(existing_files['all'])))
    files = []
    log.info(f"Adding {len(new_files)} new files to index...")
    for path in new_files:
        relpath = os.path.relpath(path, hkcfg.hk_root)
        files.append(HkFile(
            path=relpath,
            size=os.path.getsize(path),
            mod_time=os.path.getmtime(path),
            index_status='unindexed'
        ))

    session.add_all(files)
    session.commit()

    return {
        'scanned_subdirs': len(subdirs),
        'total_files': len(all_files),
        'new_files': len(new_files),
    }


def get_frames_from_file(
    hkcfg: HkConfig,
    file: HkFile,
    return_on_fail=True
) -> List[HkFrame]:
    """
    Returns HkFile and HkFrame objects corresponding to a given hk file.

    Args
    --------
    file : HkFile
        HkFile object corresponding to the file
    return_on_fail : bool
        If True, if there is a runtime error while reading the g3 file (usually
        caused by a forced shutdown), the function will still return parsed
        frames.

    Returns
    ---------
    frames : List[HkFrame]
        List of all HkFrames in the file
    """
    frames = []
    if os.path.isabs(file.path):
        path = str(file.path)
    else:
        path = os.path.join(hkcfg.hk_root, str(file.path))
    reader = so3g.G3IndexedReader(path)

    while True:
        byte_offset = reader.Tell()
        try:
            frame = reader.Process(None)
        except RuntimeError:
            log.error(f"Error processing file {file.path} byte offset: {byte_offset}")
            if return_on_fail:
                break
            else:
                raise
        if not frame:
            break
        else:
            frame = frame[0]

        if frame['hkagg_type'] != so3g.HKFrameType.data:
            continue

        # Process frame
        addr = frame['address']
        _, agent, _, feed = addr.split('.')
        start_time, stop_time = 1<<32, 0
        fields = []
        for block in frame['blocks']:
            ts = np.array(block.times) / spt3g_core.G3Units.s
            fields.extend(block.keys())
            start_time = min(start_time, ts[0].item())
            stop_time = max(stop_time, ts[-1].item())

        fields_hash = hashlib.sha256(','.join(sorted(fields)).encode('ascii')).hexdigest()[:16]
        frames.append(HkFrame(
            agent=agent, feed=feed, fields_hash=fields_hash, byte_offset=byte_offset,
            start_time=start_time, end_time=stop_time, file=file
        ))

    return frames


def update_frame_index(hkcfg: HkConfig, session=None):
    """Updates HkFrames database with frames from unindexed files"""
    if session is None:
        hkdb = HkDb(hkcfg)
        session = hkdb.Session()

    report = {
        'new_files_indexed': 0,
        'new_files_failed': 0,
    }

    files = session.query(HkFile).filter(HkFile.index_status == 'unindexed').all()
    log.info(f"Indexing {len(files)} files")
    for file in tqdm(files, disable=(not hkcfg.show_index_pb), ascii=True):
        frames = get_frames_from_file(hkcfg, file)
        file_start, file_end = 1<<32, 0
        for f in frames:
            file_start = min(file_start, f.start_time)
            file_end = max(file_end, f.end_time)
        file.start_time = file_start
        file.end_time = file_end
        file.index_status ='indexed'
        try:
            session.add_all(frames)
            session.commit()
            report['new_files_indexed'] += 1
            log.debug(f"Committed index for {file.path}.")
        except Exception as e:
            log.error(f"Failed to commit frames from {file.path}.")
            session.rollback()
            file.index_status = 'failed'
            session.commit()
            report['new_files_failed'] += 1

    return report


def update_index_all(cfg: Union[HkConfig, str], subdirs=None,
                     retry_failed: bool=False):
    """Updates all HK index databases, and returns a report."""
    if isinstance(cfg, str):
        cfg = HkConfig.from_yaml(cfg)
    hkdb = HkDb(cfg)
    session = hkdb.Session()
    report1 = update_file_index(cfg, session=session, subdirs=subdirs,
                                retry_failed=retry_failed)
    report2 = update_frame_index(cfg, session=session)
    return report1 | report2


def purge_unindexed_files(hkcfg: HkConfig):
    """Remove any 'unindexed' files from the database.  This can be used
    to ignore files that disappeared from filesystem before being indexed.

    """
    hkdb = HkDb(hkcfg)
    session = hkdb.Session()
    session.query(HkFile).filter(HkFile.index_status == 'unindexed').delete(
        synchronize_session=False)
    session.commit()


def reset_failed_files(hkcfg: HkConfig, pattern=None):
    """Reset "failed" files to the "unindexed" state.  If pattern is
    specified, it should be an sql-compatible matching string, which
    will be matched against the path. Returns the number of changed
    rows.

    """
    hkdb = HkDb(hkcfg)
    session = hkdb.Session()
    q = session.query(HkFile) \
               .filter(HkFile.index_status == 'failed')
    if pattern is not None:
        q = q.filter(HkFile.path.like(pattern))
    rowcount = q.update({HkFile.index_status: 'unindexed'},
                 synchronize_session=False)
    session.commit()
    return rowcount


def get_files_info(hkcfg: HkConfig, index_status=None, limit=None):
    """Query file rows; sorted by path.  Optionally filter by index_status
    (list of str) and limit (int).  Returns list of HkFile.

    """
    hkdb = HkDb(hkcfg)
    session = hkdb.Session()
    q = session.query(HkFile)
    if index_status is not None:
        if isinstance(index_status, str):
            index_status = [index_status]
        q = q.filter(HkFile.index_status.in_(index_status))
    q = q.order_by(HkFile.path)
    if limit:
        q = q.limit(limit)
    rows = q.all()
    session.close()
    return rows


#####################
# HK Loading stuff
#####################
@dataclass
class Field:
    agent: str
    feed: str
    field: str

    def __str__(self):
        return f"{self.agent}.{self.feed}.{self.field}"

    @staticmethod
    def _wcmatch(a, b):
        if '*' in a:
            return fnmatch.fnmatch(b, a)
        if '*' in b:
            return fnmatch.fnmatch(a, b)
        return a == b

    def matches(self, other):
        return (self.agent == other.agent
                and Field._wcmatch(self.feed, other.feed)
                and Field._wcmatch(self.field, other.field))

    @classmethod
    def from_str(cls, s):
        try:
            agent, feed, field = s.split('.')
        except Exception:
            raise ValueError(f"Could not parse field: {s}")

        return cls(agent, feed, field)

@dataclass
class LoadSpec:
    """
    HK loading specification

    Args
    -----
    cfg: HkConfig
        Configuration object
    fields: List[str]
        List of field specifications to load. This can either be a field
        descriptor, of the format ``agent.feed.field``, or an alias defined in
        the config. Field descriptors can contain wildcards in the feed and
        field portion., for instance ``agent.*.*`` will load all fields
        belonging to the specified agent.  ``agent.feed.*`` and
        ``agent.*.*word*`` will also work as expected.
    start: float
        Start time to load
    end: float
        End time to load
    downsample_factor: int
        Downsample factor for data
    hkdb: Optional[HkDb]
        HkDb instance to use. If not specified, will create a new one from the
        cfg. This should be set manually if you are calling ``load_hk`` in a loop
        to prevent connection build-up.
    """
    cfg: HkConfig
    fields: List[str]
    start: float
    end: float
    downsample_factor: int = 1
    hkdb: Optional[HkDb] = None

    def __post_init__(self):
        fs = []
        for f in self.fields:
            if f in self.cfg.aliases:
                fs.append(Field.from_str(self.cfg.aliases[f]))
            else:
                fs.append(Field.from_str(f))
        self.fields = fs


class HkResult:
    """
    Helper class for storing results of LoadHk. If aliases are set for any
    of the keys, they will be set as attributes of the object.

    Attributes
    ------------
    data: dict
        Dict where the key is the field descriptor, and the value is a list
        where val[0] are timestamps, and val[1] is data.
    """
    def __init__(self, data, aliases=None):
        if aliases is None:
            aliases = {}
        self._aliases = aliases
        self.data = data
        for alias, key in aliases.items():
            if key in self.data:
                setattr(self, alias, self.data[key])

    def save(self, path):
        np.savez(path, aliases=np.array(self._aliases), **self.data)

    @classmethod
    def load(cls, path):
        d = np.load(path, allow_pickle=True)
        aliases = d['aliases'].item()
        data = {k: d[k] for k in d.files if k != 'aliases'}
        return cls(data, aliases=aliases)


def load_hk(load_spec: Union[LoadSpec, dict], show_pb=False,
            fields=None, start=None, end=None,
            _field_list_scan=False):
    """
    Loads hk data

    Args
    ------
    load_spec: LoadSpec
        Load specification. See docstrings of the LoadSpec class.
    show_pb: bool
        If true, will show a progressbar :)
    fields: List[str]
        Fields to load (overrides load_spec.fields).
    start: float
        Starting timestamp (overrides load_spec.start).
    end: float
        Ending timestamp (overrides load_spec.end).
    _field_list_scan: bool
        Run in special mode to support get_field_list.
    """
    if isinstance(load_spec, dict):
        load_spec = LoadSpec(**load_spec)

    if load_spec.hkdb is not None:
        hkdb: HkDb = load_spec.hkdb
    else:
        hkdb = HkDb(load_spec.cfg)

    if fields is None:
        fields = load_spec.fields
    fields = [Field.from_str(f) if isinstance(f, str) else f
              for f in fields]

    if start is None:
        start = load_spec.start
    if end is None:
        end = load_spec.end

    agent_set = list(set(f.agent for f in fields))

    file_spec = {}  # {path: [offsets]}
    feeds = []
    with hkdb.Session.begin() as sess:
        query = sess.query(HkFrame).filter(
            HkFrame.start_time <= end,
            HkFrame.end_time >= start,
            HkFrame.agent.in_(agent_set)
        ).order_by(HkFrame.start_time)
        for frame in query:
            if _field_list_scan:
                feed_key = (frame.agent, frame.feed, frame.fields_hash)
                if feed_key in feeds:
                    continue
                feeds.append(feed_key)
            if frame.file.path not in file_spec:
                file_spec[frame.file.path] = []
            file_spec[frame.file.path].append(frame.byte_offset)

    # Convert all paths to absolute paths based on cfg.hk_root
    def create_abs_path(path):
        if os.path.isabs(path):
            return path
        return os.path.join(load_spec.cfg.hk_root, path)
    file_spec = {create_abs_path(k): v for k, v in file_spec.items()}

    result = {}  # {field: [timestamps, data]}
    field_misses = set()
    def get_result_field(agent, feed, field_name):
        f = Field(agent, feed, field_name)
        key = str(f)

        if key in result:
            return result[key]

        if key in field_misses:
            return None

        for field in fields:
            if field.matches(f):
                result[key] = [[], []]
                return result[key]
        # Cache field on miss
        field_misses.add(key)
        return None

    ds_factor = load_spec.downsample_factor
    nframes = np.sum([len(offsets) for offsets in file_spec.values()])

    pb = tqdm(total=nframes, disable=(not show_pb))
    for path in sorted(list(file_spec.keys())):
        offsets = file_spec[path]
        reader = so3g.G3IndexedReader(path)
        for offset in sorted(offsets):
            reader.Seek(offset)
            frame = reader.Process(None)[0]
            addr = frame['address']
            _, agent, _, feed = addr.split('.')
            for block in frame['blocks']:
                ts = np.array(block.times)[::ds_factor] / spt3g_core.G3Units.s
                for field_name, data in block.items():
                    field = get_result_field(agent, feed, field_name)
                    if field is None or _field_list_scan:
                        continue
                    field[0].append(ts)
                    field[1].append(np.array(data)[::ds_factor])
            pb.update()
    pb.close()

    if _field_list_scan:
        return list(result.keys())

    for k, d in result.items():
        if len(d[0]) == 0:
            result[k] = (np.array([]), np.array([]))
        else:
            result[k] = (np.hstack(d[0]), np.hstack(d[1]))

    return HkResult(result, aliases=load_spec.cfg.aliases)


def get_feed_list(load_spec: Union[LoadSpec, dict]) -> List[str]:
    """Return the list of feeds present in the db and for the time range
    specified by a LoadSpec.

    Args
    ----
    load_spec: LoadSpec
        Load specification. See docstrings of the LoadSpec class.

    Returns
    -------
    List[str]
      The list of feeds, as field spec strings, with wildcard for the
      field e.g. "an_agent.a_feed.*".

    Notes
    -----
    The .start_time and .end_time are respected in the query, but the
    .fields entry is ignored in the search.

    """
    if isinstance(load_spec, dict):
        load_spec = LoadSpec(**load_spec)

    if load_spec.hkdb is not None:
        hkdb: HkDb = load_spec.hkdb
    else:
        hkdb = HkDb(load_spec.cfg)

    pairs = set()
    with hkdb.Session.begin() as sess:
        query = sess.query(HkFrame.agent, HkFrame.feed).filter(
            HkFrame.start_time <= load_spec.end,
            HkFrame.end_time >= load_spec.start,
        )
        for row in query:
            pairs.add(tuple(row))

    return [f'{a}.{b}.*' for a, b in sorted(list(pairs))]


def get_field_list(load_spec: Union[LoadSpec, dict],
                   fields: List[Field]=None) -> List[str]:
    """Inspect the HK files to get the field names associated with
    each feed covered by the load_spec (or by the fields argument).
    This is shallow search in that only a single frame from every
    ``agent.feed`` combination matching the fields list is inspected.

    Args
    ----
    load_spec: LoadSpec
        Load specification. See docstrings of the LoadSpec class.
    fields:
        List of fields (which may include wildcards) to match against.

    Returns
    -------
    List[str]
      The list of fields, as field spec strings.

    Notes
    -----

    If fields is not specified, then it is taken from load_spec.  But
    normally you'd want it from the get_feeds_list.  The .start_time
    and .end_time are respected in the query, especially in the sense
    that the shallow data search will begin at .start_time.

    """
    return load_hk(load_spec, fields=fields, _field_list_scan=True)



import numpy as np


def detrend_tod(
    tod,
    method="linear",
    axis_name="samps",
    signal_name="signal",
    in_place=True,
    wrap_name=None,
    count=10,
):
    """Returns detrended data. Detrends data in place by default but pass
    in_place=False if you would like a copied array (such as if you're just
    looking to use this in an FFT).

    Using this with method ='mean' and axis_name='dets' will remove a
    common mode from the detectors
    Using this with method ='median' and axis_name='dets' will remove a
    common mode from the detectors with the median rather than the mean

    Arguments
    ---------
        tod: axis manager
        method: str
            method of detrending can be 'linear', 'mean', or median
        axis_name: str
            the axis along which to detrend. default is 'samps'
        signal_name: str
            the name of the signal to detrend. defaults to 'signal'. Can have
            any shape as long as axis_name can be resolved.
        in_place: bool.
            If False it makes a copy of signal before detrending and returns
            the copy.
        wrap_name: str or None.
            If not None, wrap the detrended data into tod with this name.
        count: int
            Number of samples to use, on each end, when measuring mean level
            for 'linear' detrend.  Values larger than 1 suppress the influence
            of white noise.

    Returns
    -------
        signal: array of type tod[signal_name]
            Detrended signal. Done in place or on a copy depend on in_place
            argument.
    """

    signal = tod[signal_name]
    if not in_place:
        signal = signal.copy()
    dtype_in = signal.dtype

    axis_idx = list(tod._assignments[signal_name]).index(axis_name)
    n_samps = signal.shape[axis_idx]

    # Ensure last axis is the one to detrend.
    # note that any axis re-ordering is not the slow part of this function
    if axis_idx != signal.ndim - 1:
        # Note this reordering is its own inverse; e.g. [0, 1, -1, 3, 4, 2]
        axis_reorder = list(range(signal.ndim))
        axis_reorder[axis_idx], axis_reorder[-1] = -1, axis_idx
        signal = signal.transpose(tuple(axis_reorder))

    if method == "mean":
        signal -= np.mean(signal, axis=-1)[..., None]
    elif method == "median":
        signal -= np.median(signal, axis=-1)[..., None]
    elif method == "linear":
        x = np.linspace(0, 1, n_samps, dtype=dtype_in)
        count = max(1, min(count, signal.shape[-1] // 2))
        slopes = signal[..., -count:].mean(axis=-1, dtype=dtype_in) - signal[
            ..., :count
        ].mean(axis=-1, dtype=dtype_in)

        # the 2d loop is significantly faster if possible
        if len(signal.shape) == 2:
            for i in range(signal.shape[0]):
                signal[i, :] -= slopes[i] * x
        else:
            signal -= slopes[..., None] * x
        signal -= np.mean(signal, axis=-1)[..., None]
    else:
        raise ValueError("method flag must be linear, mean, or median")

    if axis_idx != signal.ndim - 1:
        signal = signal.transpose(tuple(axis_reorder))

    assert signal.dtype == dtype_in

    if wrap_name is not None:
        axis_map = [(i, x) for i, x in enumerate(tod._assignments[signal_name])]
        tod.wrap(wrap_name, signal, axis_map)

    return signal



""" This module includes all the database definitions used with G3tSmurf, broken out 
here to reduce the overall number of lines in each module. 
"""
from enum import Enum
import sqlalchemy as db
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker, relationship, backref

Base = declarative_base()

association_table = db.Table(
    "association_chan_assign",
    Base.metadata,
    db.Column("tunesets", db.Integer, db.ForeignKey("tunesets.id")),
    db.Column("chan_assignments", db.Integer, db.ForeignKey("chan_assignments.id")),
)

association_table_obs = db.Table(
    "association_obs",
    Base.metadata,
    db.Column("tunesets", db.Integer, db.ForeignKey("tunesets.id")),
    db.Column("observations", db.Integer, db.ForeignKey("obs.obs_id")),
)

association_table_dets = db.Table(
    "detsets",
    Base.metadata,
    db.Column("name", db.Integer, db.ForeignKey("tunesets.name")),
    db.Column("det", db.Integer, db.ForeignKey("channels.name")),
)


class TimeCodes(Base):
    """Table for tracking if files and metadata are ready for bookbinding. Built
    to work off of the suprsync finalization files.

    Attributes
    ----------
    stream_id: string
        stream_id found in the timecode files
    suprsync_type: integer
        entry in SupRsyncType class. For tracking if suprsync is managing the
        timestreams (files) or smurf (meta) files
    timecode: integer
        the 5 digit ctime folder number that has been finalized
    agent: string
        instance id of the suprsync agent we got the finalization information
        from.
    """

    __tablename__ = "time_codes"
    __table_args__ = (db.UniqueConstraint(
        "stream_id", 
        "suprsync_type", 
        "timecode",
        "agent",
    ),)
    id = db.Column(db.Integer, primary_key=True)
    stream_id = db.Column(db.String)
    suprsync_type = db.Column(db.Integer)
    timecode = db.Column(db.Integer)
    agent = db.Column(db.String)


class SupRsyncType(Enum):
    """Files match with the timestream folders and meta match with the smurf
    folders.
    """

    FILES = 0
    META = 1

    @classmethod
    def from_string(cls, string):
        if string == "smurf":
            return cls.META
        elif string == "timestreams":
            return cls.FILES
        else:
            raise ValueError("SupRsync strings are 'smurf' or 'timestreams'")


class Finalize(Base):
    """Table for tracking what the finalization times for different agents were
    at the last g3tsmurf update. Will have a special row for the archive update
    time that will have the name G3tSMURF.

    Attributes
    -----------
    agent: string
        instance id of the suprsync agents
    time: float
        the finalization timestamp of the agent at the last update
    """

    __tablename__ = "finalize"
    id = db.Column(db.Integer, primary_key=True)
    agent = db.Column(db.String, unique=True)
    time = db.Column(db.Float)


class Observations(Base):
    """Times of continuous detector readout. This table is named obs and serves
    as the ObsDb table when loading via Context. This table is meant to by built
    off of Level 2 data, before the data from different stream_id/smurf slots have
    been bookbound and perfectly co-sampled.

    Observations are not created if the action folder has no associated .g3 files.

    Dec. 2021 -- The definitions of obs_id and timestamp changed to better
    match the operation of the smurf-streamer / sodetlib / pysmurf.
    Oct. 2022 -- The definition of obs_id changed again to include obs or oper
    tags based on if the observation is an sodetlib operation or not.

    Attributes
    -----------
    obs_id : string
        <obs|oper>_<stream_id>_<session_id>.
    timestamp : integer
        The .g3 session_id, which is also the ctime the .g3 streaming started
        and the first part .g3 file name.
    action_ctime : integer
        The ctime of the pysmurf action, generally slightly different than
        the .g3 session_id
    action_name : stream
        The name of the action used to create the observation.
    stream_id : string
        The stream_id of this observation. Generally corresponds to UFM or Smurf
        slot. Column is implemented since level 2 data is not perfectly co-sampled
        across stream_ids.
    timing : bool
        If true, the files of the entry observation were made with times
        aligned to the external timing system and high precision timestamps.
    duration : float
        The total observation time in seconds
    n_samples : integer
        The total number of samples in the observation
    start : datetime.datetime
        The start of the observation as a datetime object
    stop : datetime.datetime
        The end of the observation as a datetime object
    tag : string
        Tags for this observation in a single comma delimited string. These are populated
        through tags set while running sodetlib's stream data functions.
    calibration : bool
        Boolean that stores whether or not the observation is a calibration-type observation
        i.e. an IV curve, a bias step, etc.
    files : list of SQLAlchemy instances of Files
        The list of .g3 files in this observation built through a relationship
        to the Files table. [f.name for f in Observation.files] will return
        absolute paths to all the files.
    tunesets : list of SQLAlchemy instances of TuneSets
        The TuneSets used in this observation. There is expected to be
        one per stream_id (SMuRF crate slot).
    """

    __tablename__ = "obs"

    obs_id = db.Column(db.String, primary_key=True)
    timestamp = db.Column(db.Integer)
    action_ctime = db.Column(db.Integer)
    action_name = db.Column(db.String)

    stream_id = db.Column(db.String)
    timing = db.Column(db.Boolean)

    # in seconds
    duration = db.Column(db.Float)
    n_samples = db.Column(db.Integer)

    start = db.Column(db.DateTime)
    stop = db.Column(db.DateTime)

    tag = db.Column(db.String)

    calibration = db.Column(db.Boolean)
    ## one to many
    files = relationship(
        "Files",
        back_populates="observation",
        order_by="Files.start",
    )

    ## many to many
    tunesets = relationship(
        "TuneSets", secondary=association_table_obs, back_populates="observations"
    )

    def __repr__(self):
        try:
            return f"{self.obs_id}: {self.start} -> {self.stop} [{self.stop-self.start}] ({self.tag})"
        except:
            return f"{self.obs_id}: {self.start} -> {self.stop} ({self.tag})"


class Tags(Base):
    """Tags used to mark Observations.

    To have these automatically added, use
    sodetlib.smurf_funct.smurf_ops.stream_g3_on( S, tag= 'tag1,tag2')
    or
    sodetlib.smurf_funct.smurf_ops.take_g3_data(S, dur, tag='tag1,tag2')

    Note, this table is not relationally mapped to the Observation.tag column.
        It is set up to match the Context design
    """

    __tablename__ = "tags"
    id = db.Column(db.Integer, primary_key=True)

    obs_id = db.Column(db.String)
    tag = db.Column(db.String)


class Files(Base):
    """Table to store file indexing info. This table is named files in sql and
    serves as the ObsFileDb when loading via Context.

    Attributes
    ------------
    id : integer
        auto-incremented primary key
    name : string
        complete absolute path to file
    start : datetime.datetime
        the start time for the file
    stop : datetime.datetime
        the stop time for the file
    sample_start : integer
        Not Implemented Yet
    sample_stop : integer
        Not Implemented Yet
    obs_id : String
        observation id linking Files table to the Observation table
    observation : SQLAlchemy Observation Instance
    stream_id : The stream_id for the file. Generally of the form crateXslotY.
        These are expected to map one per UXM.
    timing : bool
        If true, every frame in the file has high precision timestamps (slightly
        different definition than obs.timing)
    n_frames : Integer
        Number of frames in the .g3 file
    frames : list of SQLALchemy Frame Instances
        List of database entries for the frames in this file
    n_channels : Integer
        The number of channels read out in this file
    detset : string
        TuneSet.name for this file. Used to map the TuneSet table to the Files
        table. Called detset to serve duel-purpose and map files to detsets
        while loading through Context.
    tuneset : SQLAlchemy TuneSet Instance
    tune_id : integer
        id of the tune file used in this file. Used to map Tune table to the
        Files table.
    tune : SQLAlchemy Tune Instance
    """

    __tablename__ = "files"
    id = db.Column(db.Integer, primary_key=True)

    name = db.Column(db.String, nullable=False, unique=True)

    start = db.Column(db.DateTime)
    stop = db.Column(db.DateTime)

    ## This is sample in an Observation
    sample_start = db.Column(db.Integer)
    sample_stop = db.Column(db.Integer)

    ## this is a string for compatibility with sotodlib, not because it makes
    ## sense here
    obs_id = db.Column(db.String, db.ForeignKey("obs.obs_id"))
    observation = relationship("Observations", back_populates="files")

    stream_id = db.Column(db.String)
    timing = db.Column(db.Boolean)

    n_frames = db.Column(db.Integer)
    frames = relationship("Frames", back_populates="file")

    ## n_channels is a renaming of channels
    n_channels = db.Column(db.Integer)

    # breaking from linked table convention to match with obsfiledb requirements
    ## many to one
    detset = db.Column(db.String, db.ForeignKey("tunesets.name"))
    tuneset = relationship("TuneSets", back_populates="files")

    tune_id = db.Column(db.Integer, db.ForeignKey("tunes.id"))
    tune = relationship("Tunes", back_populates="files")


class Tunes(Base):
    """Indexing of 'tunes' available during observations. Should
    correspond to all tune files.

    Attributes
    -----------
    id : integer
        primary key
    name : string
        name of tune file
    path : string
        absolute path of tune file
    stream_id : string
        stream_id for file
    start : datetime.datetime
        The time the tune file is made
    files : list of SQLAlchemy File instances
        All file using this tune file
    tuneset_id : integer
        id of tuneset this tune belongs to. Used to link Tunes table to TuneSets
        table
    tuneset : SQLAlchemy TuneSet instance
    """

    __tablename__ = "tunes"
    __table_args__ = (db.UniqueConstraint("name", "stream_id"),)
    id = db.Column(db.Integer, primary_key=True)

    name = db.Column(db.String)
    path = db.Column(db.String)
    stream_id = db.Column(db.String)

    ## should stop exist? tune file use does not need to be continguous
    start = db.Column(db.DateTime)

    ## files that use this tune file
    ## one to many
    files = relationship("Files", back_populates="tune")

    ## one to many
    tuneset_id = db.Column(db.Integer, db.ForeignKey("tunesets.id"))
    tuneset = relationship("TuneSets", back_populates="tunes")

    @staticmethod
    def get_name_from_status(status):
        """Return the name format expected from SmurfStatus instance"""
        return status.stream_id + "_" + status.tune.strip(".npy")


class TuneSets(Base):
    """Indexing of 'tunes sets' available during observations. Should
    correspond to the tune files where new_master_assignment=True. TuneSets
    exist to combine sets of <=8 Channel Assignments since SMuRF tuning and
    setup is run "per smurf band" while channel readout reads all tuned bands.
    Every TuneSet is a Tune file but not all Tunes are a Tuneset. This is
    because new Tunes can be made for the same set of channel assignments as the
    cryostat / readout environments evolve.

    Attributes
    ----------
    id : integer
        primary key
    name : string
        name of tune file
    path : string
        absolute path of tune file
    stream_id : string
        stream_id for file
    start : datetime.datetime
        The time the tune file is made
    stop : datetime.datetine
        Not Implemented Yet
    files : list of SQLAlchemy File instances.
    tunes : list of Tunes that are part of the TuneSet
    observations : list of observations that have this TuneSet
    chan_assignments : list of Channel Assignments that are part of the this
        tuneset
    channels : list of Channels that are part of this TuneSet
    """

    __tablename__ = "tunesets"
    __table_args__ = (db.UniqueConstraint("name", "stream_id"),)
    id = db.Column(db.Integer, primary_key=True)

    name = db.Column(db.String)
    path = db.Column(db.String)
    stream_id = db.Column(db.String)

    ## should stop exist? tune file use does not need to be continguous
    start = db.Column(db.DateTime)
    stop = db.Column(db.DateTime)

    ## files that use this detset
    ## one to many
    files = relationship("Files", back_populates="tuneset")
    tunes = relationship("Tunes", back_populates="tuneset")
    ## many to many
    observations = relationship(
        "Observations", secondary=association_table_obs, back_populates="tunesets"
    )

    ## many to many
    chan_assignments = relationship(
        "ChanAssignments", secondary=association_table, back_populates="tunesets"
    )

    ## many to many
    dets = relationship(
        "Channels", secondary=association_table_dets, back_populates="detsets"
    )

    @property
    def channels(self):
        return self.dets


class ChanAssignments(Base):
    """The available channel assignments. TuneSets are made of up to eight of
    these assignments.

    Attributes
    ----------
    id : integer
        primary key
    ctime : integer
        ctime where the channel assignment was made
    name : string
        name of the channel assignment file
    path : string
        absolute path of channel assignment file
    stream_id : string
        stream_id for file
    band : integer
        Smurf band for this channel assignment
    tunesets : list of SQLAlchemy tunesets
        The Tunesets the channel assignment is in
    channels : list of SQLAlchemy channels
        The channels in the channel assignment
    """

    __tablename__ = "chan_assignments"
    id = db.Column(db.Integer, primary_key=True)

    ctime = db.Column(db.Integer)
    path = db.Column(db.String, unique=True)
    name = db.Column(db.String)
    stream_id = db.Column(db.String)
    band = db.Column(db.Integer)

    ## Channel Assignments are put into detector sets
    ## many to many bidirectional
    tunesets = relationship(
        "TuneSets", secondary=association_table, back_populates="chan_assignments"
    )

    ## Each channel assignment is made of many channels
    ## one to many
    channels = relationship("Channels", back_populates="chan_assignment")


class Channels(Base):
    """All the channels tracked by SMuRF indexed by the ctime of the channel
    assignment file, SMuRF band and channel number. Many channels will map to
    one detector on a UFM.

    Dec. 2021 -- Updated channel names to include stream_id to ensure uniqueness

    Attributes
    ----------
    id : integer
        primary key
    name : string
        name of of channel. This is the unique readout id that will be matched with
        the unique detector id. Has the form of sch_<stream_id>_<ctime>_<band>_<channel>
    stream_id : string
        stream_id for file
    subband : integer
        The subband of the channel
    channel : integer
        The assigned smurf channel
    frequency : float
        The frequency of the resonator when the channel assigments were made
    band : integer
        The smurf band for the channel
    ca_id : integer
        The id of the channel assignment for the channel. Used for SQL mapping
    chan_assignment : SQLAlchemy ChanAssignments Instance
    detsets : List of SQLAlchemy Tunesets
        The tunesets the channel can be found in
    """

    __tablename__ = "channels"
    id = db.Column(db.Integer, primary_key=True)
    name = db.Column(db.String, unique=True)
    stream_id = db.Column(db.String)

    ## smurf channels
    subband = db.Column(db.Integer)
    channel = db.Column(db.Integer)
    frequency = db.Column(db.Float)

    band = db.Column(db.Integer)

    ## many to one
    ca_id = db.Column(db.Integer, db.ForeignKey("chan_assignments.id"))
    chan_assignment = relationship("ChanAssignments", back_populates="channels")

    ## many to many
    detsets = relationship(
        "TuneSets", secondary=association_table_dets, back_populates="dets"
    )


class FrameType(Base):
    """Enum table for storing frame types"""

    __tablename__ = "frame_type"
    id = db.Column(db.Integer, primary_key=True)
    type_name = db.Column(db.String, unique=True, nullable=True)


class Frames(Base):
    """Table to store frame indexing info
    Attributes
    ----------
    id : Integer
        Primary Key
    file_id : integer
        id of the file the frame is part of, used for SQL mapping
    file : SQLAlchemy instance
    frame_idx : integer
        frame index in the file
    offset : integer
        frame offset used by so3g indexed reader
    type_name : string
        frame types, use Observation, Wiring, and Scan frames
    time : datetime.datetime
        The start of the frame
    n_samples : integer
        the number of samples in the frame
    n_channels : integer
        the number of channels in the frame
    start : datetime.datetime
        the start of the frame
    stop : datetime.datetime
        the end of the frame
    """

    __tablename__ = "frame_offsets"
    __table_args__ = (db.UniqueConstraint("file_id", "frame_idx", name="_frame_loc"),)

    id = db.Column(db.Integer, primary_key=True)

    file_id = db.Column(db.Integer, db.ForeignKey("files.id"))
    file = relationship("Files", back_populates="frames")

    frame_idx = db.Column(db.Integer, nullable=False)
    offset = db.Column(db.Integer, nullable=False)

    type_name = db.Column(db.String, db.ForeignKey("frame_type.type_name"))
    frame_type = relationship("FrameType")

    status_dump = db.Column(db.Boolean, nullable=False, default=False)
    time = db.Column(db.DateTime, nullable=False)

    # Specific to data frames
    n_samples = db.Column(db.Integer)
    n_channels = db.Column(db.Integer)
    start = db.Column(db.DateTime)
    stop = db.Column(db.DateTime)

    def __repr__(self):
        return f"Frame({self.type_name})<{self.frame_idx}>"



# Copyright (c) 2018-2024 Simons Observatory.
# Full license can be found in the top level "LICENSE" file.

import os
import pickle

import ephem
import h5py
import healpy as hp
import numpy as np
import traitlets
from astropy import constants
from astropy import units as u
from scipy.constants import au as AU
from scipy.interpolate import RectBivariateSpline
from toast import qarray as qa
from toast.data import Data
from toast.observation import default_values as defaults
from toast.ops.operator import Operator
from toast.timing import function_timer, Timer
from toast.traits import (Bool, Float, Instance, Int, Quantity, Unicode,
                          trait_docs)
from toast.utils import Environment, Logger, unit_conversion

XAXIS, YAXIS, ZAXIS = np.eye(3)


@trait_docs
class SimStimulator(Operator):
    """Simulate stimulator calibration signal"""

    API = Int(0, help="Internal interface version for this operator")

    times = Unicode(defaults.times, help="Observation shared key for timestamps")

    shared_flags = Unicode(
        defaults.shared_flags,
        allow_none=True,
        help="Observation shared key for telescope flags to use",
    )

    shared_flag_mask_stimulator = Int(
        defaults.shared_mask_invalid, help="Bit mask value for optional shared flagging"
    )

    shared_flag_mask_unstable = Int(
        255, help="Bit mask value applied during stimulator calibration"
    )

    chopper_rates = Unicode(
        "7,16,35,65,100,125,150",
        help="Chopper modulation rates as comma-separated values [Hz]",
    )

    chopper_blade_count = Int(4, help="Number of chopper blades")

    chopper_blade_width = Quantity(45 * u.degree, help="Chopper blade_width")

    chopper_acceleration = Quantity(
        300 * u.radian / u.s ** 2, help="Angular acceleration of the chopper"
    )

    chopper_step_time = Quantity(60 * u.s, help="Single chopper frequency step length")

    heater_temperature = Quantity(77 * u.mK, help="Stimulator target temperature")

    heater_aperture_diameter = Quantity(44 * u.mm, help="Heater aperture at chopper")

    heater_aperture_distance = Quantity(70.5 * u.mm, help="Heater aperture from chopper axis")

    blackbody_temperature = Quantity(17 * u.mK, help="Chopper temperature")

    chopper_state = Unicode("chopper_state", help="Observation key for chopper state")

    det_data = Unicode(
        defaults.det_data,
        help="Observation detdata key for simulated signal",
    )

    stimulator_temperature = Unicode(
        "stimulator", help="Observation shared key for stimulator temperature"
    )

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    @function_timer
    def _get_chopper_state(self, obs):
        """Chopper modulates the stimulator signal"""

        my_times = obs.shared[self.times].data

        comm = obs.comm_row
        if comm is None or comm.size == 1:
            times_tot = my_times
        else:
            times_tot = np.hstack(comm.allgather(my_times))

        # Evaluate the fraction of heater aperture that is visible
        # as a function of chopper angle

        d = self.heater_aperture_distance.to_value(u.m)
        R = self.heater_aperture_diameter.to_value(u.m) / 2
        D = np.sqrt(d ** 2 - R ** 2)
        period = 2 * np.pi / self.chopper_blade_count

        # Half of the aperture angle measured from the chopper spin axis
        alpha = np.arcsin(R / d)

        # Chopper blade width
        gamma = self.chopper_blade_width.to_value(u.radian)

        def get_ratio(beta):
            """Return the fraction of covered aperture.

            Beta measures the overlap between the blade and the aperture.
            Defined only for beta=[0,alpha]. Ratio(alpha)=0.5
            """
            tanbeta = np.tan(beta)
            c = 2 * np.sqrt(
                (D + R * tanbeta) ** 2 / (1 + tanbeta ** 2) - D ** 2
            )
            theta = 2 * np.arcsin(0.5 * c / R)
            ratio = 0.5 * (theta - np.sin(theta)) / np.pi
            return ratio

        # One full cycle
        all_beta = []
        all_ratio = []
        n = 1000
        eps = 1e-6  # avoid singularities

        # Blade covers first half of the aperture
        beta = np.linspace(0, alpha, n, endpoint=False)
        ratio = get_ratio(beta)
        all_beta.append(beta)
        all_ratio.append(ratio)

        # Blade covers second half of the aperture
        beta = np.linspace(alpha + eps, 2 * alpha, n, endpoint=False)
        ratio = 1 - get_ratio(2 * alpha - beta)
        all_beta.append(beta)
        all_ratio.append(ratio)

        # Blade fully covers aperture
        beta = np.linspace(2 * alpha, gamma, n, endpoint=False)
        ratio = np.ones_like(beta)
        all_beta.append(beta)
        all_ratio.append(ratio)

        # Blade uncovers first half of the aperture
        beta = np.linspace(gamma, gamma + alpha, n, endpoint=False)
        ratio = 1 - get_ratio(beta - gamma)
        all_beta.append(beta)
        all_ratio.append(ratio)

        # Blade uncovers second half of the aperture
        beta = np.linspace(gamma + alpha + eps, gamma + 2 * alpha, n, endpoint=False)
        ratio = get_ratio(gamma + 2 * alpha - beta)
        all_beta.append(beta)
        all_ratio.append(ratio)

        # No coverage
        beta = np.linspace(gamma + 2 * alpha, 2 * np.pi / self.chopper_blade_count, n)
        ratio = np.zeros_like(beta)
        all_beta.append(beta)
        all_ratio.append(ratio)

        # Concatenate
        beta = np.hstack(all_beta)
        ratio = np.hstack(all_ratio)

        # Interpolate
        # phase = np.linspace(0, 2 * np.pi, 1000)
        # illumination = np.interp(phase % period, beta, ratio)

        # Now simulate the phase

        current_phase = 0
        current_rate = 0
        current_time = times_tot[0]
        all_times = []
        all_phases = []
        unstable_periods = []
        n = 100
        rates = []
        for rate in self.chopper_rates.split(","):
            rates.append(float(rate) / self.chopper_blade_count * 2 * np.pi)
        accel = self.chopper_acceleration.to_value(u.radian / u.second ** 2)
        for rate in rates:
            # Acceleration/deceleration
            sign = np.sign(rate - current_rate)
            time_accel = sign * (rate - current_rate) / accel
            times = np.linspace(0, time_accel, n, endpoint=False)
            all_times.append(times + current_time)
            all_phases.append(
                0.5 * sign * accel * times ** 2 + current_rate * times + current_phase
            )
            unstable_periods.append([current_time, current_time + time_accel])
            current_phase += 0.5 * sign * accel * time_accel ** 2 + current_rate * time_accel
            current_time += time_accel
            current_rate = rate
            # Coasting
            time_coast = self.chopper_step_time.to_value(u.s) - time_accel
            times = np.linspace(0, time_coast, n, endpoint=False)
            all_times.append(times + current_time)
            all_phases.append(current_rate * times + current_phase)
            current_time += time_coast
            current_phase += current_rate * time_coast

        # Add last data points to support extrapolation and concatenate

        all_times.append([current_time, current_time + 1e6])
        all_phases.append([current_phase, current_phase])
        phase_times = np.hstack(all_times)
        phases = np.hstack(all_phases)

        my_phase = np.interp(my_times, phase_times, phases)
        my_state = np.interp(my_phase % period, beta, ratio)

        # Deposit in the Observation

        obs.shared.create_column(
            self.chopper_state, shape=(obs.n_local_samples,), dtype=np.float64
        )
        obs.shared[self.chopper_state].set(my_state, offset=(0,), fromrank=0)

        flags = np.zeros(obs.n_local_samples, dtype=np.uint8) \
            + self.shared_flag_mask_stimulator

        for start, stop in unstable_periods:
            ind = np.logical_and(my_times > start, my_times < stop)
            flags[ind] |= self.shared_flag_mask_unstable

        obs.shared[self.shared_flags].data |= flags

        return

    @function_timer
    def _get_stimulator_temperature(self, obs):
        """Time-dependent stimulator temperature"""
        times = obs.shared[self.times].data

        # Very simple model, constant temperature throughout the observation,

        temperature = np.ones(obs.n_local_samples) * self.heater_temperature.to_value(u.K)

        # Deposit in the Observation
        obs.shared.create_column(
            self.stimulator_temperature, shape=(obs.n_local_samples,), dtype=np.float64
        )
        obs.shared[self.stimulator_temperature].set(temperature, offset=(0,), fromrank=0)

        return

    def _get_stimulator_signal(self, obs, band, quat):

        # FIXME: figure out the actual amplitude of the signal in K_CMB

        chopper = obs.shared[self.chopper_state].data
        temperature = obs.shared[self.stimulator_temperature].data

        # This is placeholder code.  Proper calculation should account for
        # detector bandpass, location on the focalplane and possibly for
        # the chopper temperature

        signal = temperature * chopper + self.blackbody_temperature.to_value(u.K)

        return signal

    @function_timer
    def _exec(self, data, detectors=None, **kwargs):
        log = Logger.get()

        for trait in "shared_flags", :
            value = getattr(self, trait)
            if value is None:
                raise RuntimeError(f"You must set `{trait}` before running SimStimulator")

        for obs in data.obs:
            if "STIMULATOR" not in obs.name.upper():
                log.debug_rank(
                    f"SimStimulator: {obs.name} is not a stimulator calibration observation.",
                    comm=obs.comm.comm_group,
                )
                continue
            self._get_chopper_state(obs)
            self._get_stimulator_temperature(obs)

            dets = obs.select_local_detectors(detectors)
            obs.detdata.ensure(self.det_data, detectors=dets, create_units=u.K)
            det_units = obs.detdata[self.det_data].units
            scale = unit_conversion(u.K, det_units)

            focalplane = obs.telescope.focalplane
            for det in dets:
                signal = obs.detdata[self.det_data][det]
                band = focalplane[det]["band"]
                quat = focalplane[det]["quat"]

                stim = self._get_stimulator_signal(obs, band, quat)

                signal += scale * stim

        return

    def _finalize(self, data, **kwargs):
        return

    def _requires(self):
        req = self.detector_pointing.requires()
        req["shared"].append(self.times)
        return req

    def _provides(self):
        prov = {
            "meta": list(),
            "shared": [self.chopper_state, self.stimulator_temperature],
            "detdata": [self.det_data],
        }
        return prov

    def _accelerators(self):
        return list()



import unittest
from sotodlib.core import metadata

import os
import time

from ._helpers import mpi_multi


def get_example(stuff_missing=False):
    # Create a new Db and add two columns.
    obsdb = metadata.ObsDb()
    obsdb.add_obs_columns(['timestamp float', 'hwp_speed float', 'drift string'])

    # Add 10 rows.
    for i in range(10):
        if stuff_missing and i in [2, 4]:
            continue
        tags = []
        if i == 6:
            tags.append('cryo_problem')
        if i > 7:
            tags.append('planet')
        elif not stuff_missing:
            tags.append('cmb_survey')
        obsdb.update_obs(f'myobs{i}', {'timestamp': 1900000000. + i * 100,
                                       'hwp_speed': 2.0,
                                       'drift': 'rising' if i%2 else 'setting'},
                         tags=tags)
    return obsdb


@unittest.skipIf(mpi_multi(), "Running with multiple MPI processes")
class TestObsDb(unittest.TestCase):

    def setUp(self):
        pass

    def test_smoke(self):
        """Basic functionality."""
        db = get_example()
        all_obs = db.query()
        db.get(all_obs[0]['obs_id'])
        db.query('timestamp > 0')
        db.query('timestamp > 0', tags=['cryo_problem=1'])

    def test_query(self):
        db = get_example()
        r0 = db.query("drift == 'rising'")
        r1 = db.query("drift == 'setting'")
        self.assertGreater(len(r0), 0)
        self.assertEqual(len(r0) + len(r1), len(db))

        # Grumbles about double quotes?
        with self.assertWarns(UserWarning):
            r1 = db.query('drift == "setting"')

    def test_tags(self):
        db = get_example()
        r0 = db.query(tags=['planet=1', 'cryo_problem', 'not_a_tag'])
        r1 = db.query(tags=['planet=0', 'cryo_problem', 'not_a_tag'])
        self.assertGreater(len(r0), 0)
        self.assertEqual(len(r0) + len(r1), len(db))
        for k in ['planet', 'cryo_problem', 'not_a_tag']:
            self.assertTrue(k in r0.keys)
            self.assertTrue(k in r1.keys)

    def test_io(self):
        """Check to_file and from_file."""
        db0 = get_example()
        dump_list = [(f'test.sqlite', None),
                     (f'test.txt', 'dump'),
                     (f'test.gz', None)]
        # Save.
        for fn, fmt in dump_list:
            print(f'Writing {fn}')
            db0.to_file(fn, fmt=fmt)
            print('  -- output has size {}'.format(os.path.getsize(fn)))
            t0 = time.time()
            db1 = metadata.ObsDb.from_file(fn, fmt=fmt)
            dt = time.time() - t0
            self.assertEqual(len(db1.query()), len(db0.query()))
            print('  -- removing.')
            os.remove(fn)

    def test_info(self):
        """Check the .info method."""
        db0 = get_example()
        db0.info()

    def test_diff_patch(self):
        """Use diff/patch to update one obsdb to match another."""
        db0 = get_example(stuff_missing=True)
        db1 = get_example()
        diff = metadata.obsdb.diff_obsdbs(db0, db1)
        assert diff['different']
        assert diff['patchable']

        metadata.obsdb.patch_obsdb(diff['patch_data'], db0)
        diff = metadata.obsdb.diff_obsdbs(db0, db1)
        assert not diff['different']


if __name__ == '__main__':
    unittest.main()



# Copyright (c) 2021 Simons Observatory.
# Full license can be found in the top level "LICENSE" file.

"""Check tod_ops routines.

"""

import unittest
import numpy as np
import pylab as pl
import scipy.signal

from sotodlib import core, tod_ops
from sotodlib.preprocess.pcore import _expand

from numpy.testing import assert_array_equal


from ._helpers import mpi_multi
from .test_tod_ops import get_tod

from so3g.proj import Ranges, RangesMatrix
from scipy.sparse import csr_array

class TestExpand(unittest.TestCase):

    def test_100_expand(self):
        aman = get_tod(sig_type='trendy',ndets=5, nsamps=1000)
        full = core.AxisManager( aman.dets, aman.samps)
        proc_aman = core.AxisManager( aman.dets, aman.samps)
        
        # wrap a dets array with floats
        proc_aman.wrap('arr1', np.mean( aman.signal, axis=1), [(0,'dets')],) 
        # wrap a dets array with ints
        proc_aman.wrap(
            'arr2', np.round( np.min(aman.signal, axis=1)).astype('int32'),
            [(0,'dets')],
        ) 
        # wrap a dets,samps array to make sure it works 
        proc_aman.wrap('arr3', aman.signal*23, [(0,'dets'), (1,'samps')],) 

        # wrap a RangesMatrix
        proc_aman.wrap(
            'flag1', RangesMatrix.zeros( (aman.dets.count, aman.samps.count)),
            [(0,'dets'), (1,'samps')],
        ) 
        proc_aman.flag1[0].add_interval( 100,700 )
        proc_aman.flag1[3].add_interval( 110,800 )
        proc_aman.flag1[4].add_interval( 0,200 )

        # wrap a Ranges
        proc_aman.wrap('flag2', Ranges( aman.samps.count ),[(0,'samps')],) 
        proc_aman.flag2.add_interval( 450,700 )

        # wrap a csr_array
        proc_aman.wrap_new('sparse_thing', shape=('dets', 'samps'))[:] = \
            1.* (np.random.uniform(size=(aman.dets.count, aman.samps.count)) > .05)
        proc_aman.wrap('csr_thing', csr_array(proc_aman['sparse_thing']),
            [(0,'dets'), (1,'samps')])

        ## test same size expansion
        out = _expand( proc_aman, full)
        assert_array_equal( out.flag1[0].ranges()[0] , [100,700] )
        assert_array_equal( out.sparse_thing, out.csr_thing.toarray() )

        ## test with detectors cut
        aman.restrict( 'dets', aman.dets.vals[3:5])
        proc_aman.restrict( 'dets', aman.dets.vals)
        out = _expand( proc_aman, full)
        self.assertTrue( proc_aman.dets.vals[0] == out.dets.vals[3])
        assert_array_equal( out.flag1[3].ranges()[0] , [110,800] )
        assert_array_equal( out.sparse_thing, out.csr_thing.toarray() )

        aman.restrict( 'samps', (300,None))
        proc_aman.restrict( 'samps', (300,None))
        out = _expand( proc_aman, full)
        assert_array_equal( proc_aman.flag1[0].ranges()[0] , [0,500] )
        assert_array_equal( out.flag1[3].ranges()[0] , [300,800] )
        assert_array_equal( out.sparse_thing, out.csr_thing.toarray() )

        ## test with a wrapped axis manager
        dummy = core.AxisManager(
            aman.dets, 
        )
        dummy.wrap( 'blah', np.median( aman.signal, axis=1),
            [(0,'dets')],)
        proc_aman.wrap('dummy', dummy)
        out = _expand( proc_aman, full)
        
        ## check value mask
        self.assertTrue( len(out.valid[0].ranges()) == 0 )
        assert_array_equal( out.valid[3].ranges()[0], [300,1000] )

        ## test with extra axis
        pain = core.AxisManager(
            aman.dets, aman.samps, core.LabelAxis('ouch', ['x','y']),
        )
        pain.wrap_new('wammy', ('ouch', 'dets', 'samps'))
        proc_aman.wrap('pain', pain)
        out = _expand( proc_aman, full)

        ## test with no detectors
        proc_aman.restrict('dets', [])
        out = _expand( proc_aman, full)
        assert_array_equal( out.sparse_thing, out.csr_thing.toarray() )



if __name__ == '__main__':
    unittest.main()



import unittest


import numpy as np
from sotodlib import core
from sotodlib.tod_ops import flags


class TestFlags(unittest.TestCase):
    def test_trending(self):
        nsamps = 1000
        timestamps = np.arange(nsamps)
        signal = np.random.normal(size=(2, nsamps))
        signal[1] += timestamps
        dets = ["det0", "det1"]
        aman = core.AxisManager(
            core.LabelAxis("dets", dets), core.IndexAxis("samps", len(timestamps))
        )
        aman.wrap("timestamps", timestamps, [(0, "samps")])
        aman.wrap("signal", signal, [(0, "dets"), (1, "samps")])

        for max_size in (nsamps // 2, nsamps, nsamps * 2):
            cut = flags.get_trending_flags(aman, max_samples=max_size, max_trend=0.5)
            self.assertTupleEqual(cut.shape, (2, 1000))
            self.assertTrue(np.array_equal(cut.ranges[1].ranges(), [[0, 1000]]))
            self.assertEqual(len(cut.ranges[0].ranges()), 0)

        cut = flags.get_trending_flags(aman, max_trend=0.5, t_piece=333)
        self.assertTupleEqual(cut.shape, (2, 1000))
        self.assertTrue(np.array_equal(cut.ranges[1].ranges(), [[0, 1000]]))
        self.assertEqual(len(cut.ranges[0].ranges()), 0)

    def test_get_glitch_flags(self):
        nsamps = 100000
        glitch_index = 50000
        timestamps = np.arange(nsamps)
        signal = np.random.normal(0, 1, size=(2, nsamps))
        signal[1][glitch_index] = 100
        dets = ["det0", "det1"]
        aman = core.AxisManager(
            core.LabelAxis("dets", dets), core.IndexAxis("samps", len(timestamps))
        )
        aman.wrap("timestamps", timestamps, [(0, "samps")])
        aman.wrap("signal", signal, [(0, "dets"), (1, "samps")])
        aman.wrap("flags", core.AxisManager(aman.dets, aman.samps))

        # test on normal signal
        flag = flags.get_glitch_flags(aman)
        self.assertTupleEqual(flag.shape, (2, nsamps))
        self.assertEqual(len(flag.ranges[1].ranges()), 1)
        self.assertEqual(flag[1].mask()[glitch_index], True)
        self.assertEqual(len(flag.ranges[0].ranges()), 0)

        # test on 1d array such as common mode
        aman.wrap("common_mode", signal[1], [(0, "samps")])
        flag = flags.get_glitch_flags(aman, signal_name='common_mode', name='common_glitches')
        self.assertTupleEqual(flag.shape, (nsamps,))
        self.assertEqual(len(flag.ranges()), 1)
        self.assertEqual(flag.mask()[glitch_index], True)


if __name__ == "__main__":
    unittest.main()



import so3g.proj
import numpy as np
from pixell import enmap, tilemap

from .helpers import _get_csl, _valid_arg, _not_both, _confirm_wcs
from . import helpers
from . import healpix_utils as hp_utils

import logging
logger = logging.getLogger(__name__)


class P:
    """Projection Matrix.

    This class provides functions to apply a Projection Matrix (or its
    transpose).  The Projection Matrix, P, also sometimes called the
    Pointing Matrix, describes how a vector of time-ordered
    measurements d are determined by a vector of map pixels m::

      d = P m

    We are working in aspace, of course, where d is not a vector, but
    rather an array of vectors, and m is not a vector, it's a
    multi-dimensional array with two of the dimensions corresponding
    to a rectangular pixelization of the sky.

    If you are making filter+bin maps, you will want to use these functions:

    - to_map
    - to_inverse_weights
    - remove_weights

    If you are solving for a map iteratively you will probably just need:

    - to_map
    - from_map

    Important keyword arguments that are used in many functions:

    - tod: The AxisManager from which signal and pointing information
      should be taken.
    - signal: The array to use for time-ordered signal.  If a string,
      it will be looked up in tod.  Defaults to 'signal'.
    - det_weights: A vector of floats representing the inverse
      variance of each detector, under the assumption of white noise.
    - cuts: A RangesMatrix that identifies samples that should be
      excluded from projection operations.
    - comps: the component code (e.g. 'T', 'TQU', 'QU', ...) that
      specifies the spin-components being modeled in the map.
    - dest: the appropriately shaped array in which to place the
      computed result (as an alternative to allocating a new object to
      store the result).

    Note that in the case of det_weights and cuts, the Projection
    Matrix may also have cached values for those.  It is an error to
    pass either of these as a keyword argument to a projection routine
    if a value has been cached for it.

    Objects of this class cache certain pre-computed information (such
    as the rotation taking boresight coordinates to celestial
    coordinates) and certain context-dependent settings (such as a map
    shape, WCS, and spin-component configuration).  You may want to
    inspect or borrow these results, perhaps to reuse them when
    constructing new instances with slight modifications.  The cached
    attributes are:

    - sight: A CelestialSightLine, representing the boresight pointing
      in celestial coordinates. [samps]
    - fp: FocalPlane object representing the focal plane offsets
      and T,P responses of each detector. When constructing with a
      tod argument, fp can be automatically populated from
      tod.focal_plane; note that if hwp is passed as True then the
      gamma (polarization) angles will be reflected (gamma' = -gamma).
      [dets]
    - geom: The target map geometry. This is a pixell.enmap.Geometry
      object, with attributes .shape and .wcs (for non-tiled rectpix);
      a pixell.tilemap.TileGeometry (if tiled); or a Healpix geometry as
      defined in coords.healpix_utils.
    - comps: String indicating the spin-components to include in maps.
      E.g., 'T', 'QU', 'TQU'.
    - rot: quat giving an additional fixed rotation to apply to get
      from boresight to celestial coordinates.  Not for long...
    - cuts (optional): RangesMatrix indicating what samples to exclude
      from projection operations (the indicated samples have
      projection matrix element 0 in all components). [dets, samps]
    - threads (optional): list of RangesMatrix objects, to control the
      use of threads in TOD-to-map operations using OpenMP.  See so3g
      documentation. [[threads, dets, samps], ...]
    - det_weights (optional): weights (one per detector) to apply to
      time-ordered data when binning a map (and also when binning a
      weights matrix).  [dets]
    - interpol (optional): How to interpolate the values for samples
      between pixel centers. Forwarded to Projectionist. Valid
      options are:

      - None, 'nn' or 'nearest': Standard nearest neighbor mapmaking.
      - 'lin' or 'bilinear': Linearly interpolate between the four
        closest pixels.

      Default: None

    These things can be updated freely, with the following caveats:

    - If the number of "samples" or "detectors" is changed in one
      attribute, it will need to be changed in the others to match.
    - The threads attribute, if in use, needs to be recomputed if
      anything about the pointing changes (this includes map geometry
      but does not include map components).

    Setting the "threads" argument to certain special values will
    activate different thread assignment algorithms:

    - False: do not use threading; to_map projections will be
      single-threaded.
    - True: use the default algorithm, 'domdir'.
    - None: same as True.
    - 'simple': compute self.threads using simple map-stripe
      algorithm.
    - 'domdir': compute self.threads using dominant-direction
      algorithm (recommended).
    - 'tiles': for tiled geometries, design self.threads such that
      each tile is assigned to a single thread (each thread may be in
      charge of multiple tiles).

    """
    def __init__(self, sight=None, fp=None, geom=None, comps='T',
                 cuts=None, threads=None, det_weights=None, interpol=None):
        self.sight = sight
        self.fp = fp
        self.geom = wrap_geom(geom)
        self.comps = comps
        self.cuts = cuts
        self.threads = threads
        self.active_tiles = None
        self.det_weights = det_weights
        self.interpol = interpol
        self.pix_scheme = _infer_pix_scheme(self.geom)

        if self.pix_scheme == "healpix" and (self.geom.nside_tile is not None):
            if isinstance(self.geom.nside_tile, int):
                self.geom.nside_tile = min(self.geom.nside, self.geom.nside_tile)
            self.geom.ntile = True # Not used for healpix but needed for tiled()

    @classmethod
    def for_tod(cls, tod, sight=None, geom=None, comps='T',
                rot=None, cuts=None, threads=None, det_weights=None,
                timestamps=None, focal_plane=None, boresight=None,
                boresight_equ=None, wcs_kernel=None, weather='typical',
                site='so', interpol=None, hwp=False, qp_kwargs={}):
        """Set up a Projection Matrix for a TOD.  This will ultimately call
        the main P constructor, but some missing arguments will be
        extracted from tod and computed along the way.

        To determine the boresight pointing in celestial coordinates
        (ultimately passed to constructor as sight=), the first
        non-None item in the following list is used:

        - the sight= keyword argument.
        - the boresight_equ= keyword argument.
        - the boresight= keyword argument
        - tod.get('boresight_equ')
        - tod.get('boresight')

        qp_kwargs are passed through to qpoint to compute sight
        if sight and boresight_equ args are None.

        If the map geometry geom is not specified, but the wcs_kernel
        is provided, then get_footprint will be called to determine
        the geom.

        The per-detector positions and polarization directions are
        extracted from focal_plane (xi, eta, gamma), or else
        tod.focal_plane.  If hwp=True, then the rotation angles
        (gamma) are reflected (gamma' = -gamma) before storing in
        self.fp.

        """
        if sight is None:
            if boresight_equ is None:
                if boresight is None:
                    boresight_equ = tod.get('boresight_equ')
            if boresight_equ is not None:
                sight = so3g.proj.CelestialSightLine.for_lonlat(
                    boresight_equ.ra, boresight_equ.dec, boresight_equ.get('psi'))
            else:
                timestamps = _valid_arg(timestamps, 'timestamps',  src=tod)
                boresight = _valid_arg(boresight, 'boresight', src=tod)
                assert(boresight is not None)
                sight = so3g.proj.CelestialSightLine.az_el(
                    timestamps, boresight.az, boresight.el, roll=boresight.roll,
                    site=site, weather=weather, **qp_kwargs)
        else:
            sight = _get_csl(sight)

        # Apply a rotation from equatorial to map WCS coordinates.
        if rot is not None:
            sight.Q = rot * sight.Q

        # Set up the detectors in the focalplane
        fp = helpers.get_fplane(tod, focal_plane=focal_plane, hwp=hwp)

        if geom is None and wcs_kernel is not None:
            geom = helpers.get_footprint(tod, wcs_kernel, sight=sight)

        return cls(sight=sight, fp=fp, geom=geom, comps=comps,
                   cuts=cuts, threads=threads, det_weights=det_weights,
                   interpol=interpol)

    @classmethod
    def for_geom(cls, tod, geom, comps='TQU', timestamps=None,
                 focal_plane=None, boresight=None, rot=None, cuts=None):
        """Deprecated, use .for_tod."""
        return cls.for_tod(tod, geom=geom, comps=comps,
                           timestamps=timestamps, focal_plane=focal_plane,
                           boresight=boresight, rot=rot, cuts=cuts)

    def zeros(self, super_shape=None, comps=None):
        """Returns an enmap concordant with this object's configured geometry
        and component count.

        Args:
          super_shape (tuple): The leading dimensions of the array.
            If None, self._comp_count(comps) is used.
          comps: The component list, to override self.comps.

        Returns:
          An enmap with shape super_shape + self.geom.shape.

        """
        if super_shape is None:
            super_shape = (self._comp_count(comps), )
        proj, _ = self._get_proj_tiles()
        if self.pix_scheme == 'healpix':
            return proj.zeros(super_shape)
        elif self.pix_scheme == 'rectpix':
            if self.tiled:
                return tilemap.from_tiles(proj.zeros(super_shape), self.geom)
            else:
                return enmap.ndmap(proj.zeros(super_shape), wcs=self.geom.wcs)

    def to_map(self, tod=None, dest=None, comps=None, signal=None,
               det_weights=None, cuts=None, eigentol=None):
        """Project time-ordered signal into a map.  This performs the operation

            m += P d

        and returns m.

        Args:
          tod: AxisManager; possible source for 'signal', 'det_weights'.
          dest (enmap): the map or array into which the data should be
            accumulated.  (If None, a new enmap is created and
            initialized to zero.)
          signal: The time-ordered data, d.  If None, tod.signal is used.
          det_weights: The per-detector weight vector.  If None,
            self.det_weights will be used; if that is not set then
            uniform weights of 1 are applied.
          cuts: Sample cuts to exclude from processing.  If None,
            self.cuts is used.
          eigentol: This is ignored.

        """
        signal = _valid_arg(signal, 'signal', src=tod)
        det_weights = _not_both(det_weights, self.det_weights,
                                'det_weights', dtype='float32')
        cuts = _not_both(cuts, self.cuts, 'cuts')

        if comps is None: comps = self.comps
        if dest is None:  dest  = self.zeros(comps=comps)

        proj, threads = self._get_proj_threads(cuts=cuts)
        proj.to_map(signal, self._get_asm(), output=self._prepare_map(dest),
                det_weights=det_weights, comps=comps, threads=unwrap_ranges(threads))
        return dest

    def to_weights(self, tod=None, dest=None, comps=None, signal=None,
                   det_weights=None, cuts=None):
        """Computes the weights matrix for the uncorrelated noise model and
        returns it.  I.e.:

            W += P N^-1 P^T

        and returns W.  Here the inverse noise covariance has shape
        (n_dets), and carries a single weight (1/var) value for each
        detector.

        Args:
          tod (AxisManager): possible source for det_weights.
          dest (enmap): the map or array into which the weights should
            be accumulated.  (If None, a new enmap is created and
            initialized to zero.)
          det_weights: The per-detector weight vector.  If None,
            tod.det_weights will be used; if that is not set then
            uniform weights of 1 are applied.
          cuts: Sample cuts to exclude from processing.  If None,
            self.cuts is used.

        """
        det_weights = _not_both(det_weights, self.det_weights,
                                'det_weights', dtype='float32')
        cuts = _not_both(cuts, self.cuts, 'cuts')

        if comps is None:
            comps = self.comps
        if dest is None:
            _n   = self._comp_count(comps)
            dest = self.zeros((_n, _n))

        proj, threads = self._get_proj_threads(cuts=cuts)
        proj.to_weights(self._get_asm(), output=self._prepare_map(dest),
                det_weights=det_weights, comps=comps, threads=unwrap_ranges(threads))
        return dest

    def to_inverse_weights(self, weights_map=None, tod=None, dest=None,
                           comps=None, signal=None, det_weights=None, cuts=None,
                           eigentol=1e-4,
                           ):
        """Compute an inverse weights map, W^-1, from a weights map.  If no
        weights_map is passed in, it will be computed by calling
        to_weights, passing through all other arguments.

        """
        if weights_map is None:
            logger.info('to_inverse_weights: calling .to_weights')
            weights_map = self.to_weights(
                tod=tod, comps=comps, signal=signal, det_weights=det_weights, cuts=cuts)

        weights_map, uf_info = self._flatten_map(weights_map)

        if dest is not None:
            dest, uf_info = self._flatten_map(dest, uf_info)

        _dest = helpers._invert_weights_map(
                weights_map, eigentol=eigentol, UPLO='U')
        if dest is not None:
            dest[:] = _dest
        else:
            dest = _dest
        del _dest

        dest = self._unflatten_map(dest, uf_info)
        return dest

    def remove_weights(self, signal_map=None, weights_map=None, inverse_weights_map=None,
                       dest=None, **kwargs):
        """Apply the inverse weights matrix to a signal map.

              m' = W^-1 m

        If W or m are not fully specified, they will be computed by
        calling other routines inline, with relevant arguments passed
        through.

        Args:
          signal_map: The map m to filter.
          inverse_weights_map: the matrix W^-1 to apply to the map.
            Shape should be (n_comp, n_comp, n_row, n_col), but only
            the upper diagonal in the first two dimensions needs to be
            populated.  If this is None, then "weights_map" is taken
            as W, and it will be inverted and applied.
          weights_map: the matrix W.  Shape should be (n_comp, n_comp,
            n_row, n_col), but only the upper diagonal in the first
            two dimensions needs to be populated.  If this is None,
            then W will be computed and inverted via
            ``self.to_inverse_weights``.

        """
        if inverse_weights_map is None:
            inverse_weights_map = self.to_inverse_weights(weights_map=weights_map, **kwargs)
        if signal_map is None:
            signal_map = self.to_map(**kwargs)

        # Get flat numpy-compatible forms for the maps.
        signal_map, uf_info = self._flatten_map(signal_map)
        inverse_weights_map, uf_info = self._flatten_map(inverse_weights_map, uf_info)

        if dest is not None:
            dest, uf_info = self._flatten_map(dest, uf_info)

        dest = helpers._apply_inverse_weights_map(inverse_weights_map, signal_map, out=dest)
        dest = self._unflatten_map(dest, uf_info)
        return dest


    def from_map(self, signal_map, dest=None, comps=None, wrap=None,
                 cuts=None, tod=None):
        """Project from a map into the time-domain.

            d += P m

        Args:
          signal_map: The map m.  This can probably be just about
            anything supported by so3g.proj; it doesn't have to match
            the internally configured geometry.
          dest: Time-ordered data array, shape (dets, samps).  If
            None, a new array will be created to hold the result.
            Otherwise, data are *accumulated* into d, so clear it
            manually if you are trying to do d = P m.
          comps (str): Projection components, if you want to override.
          cuts: RangesMatrix, shape (dets, samps) flagging samples
            that should not be populated.  Defaults to empty.
          wrap (str): If specified, wraps the result as tod[wrap]
            (after removing whatever was in there).

        Returns:
            The dest array.

        Notes:
            Since this is a set of one-to-many operation, OpenMP can
            be used without carefully assigning samples to threads.

        """
        assert cuts is None  # whoops, not implemented.

        # This is not free but it is pretty fast, doesn't do thread
        # assignments.
        proj, _ = self._get_proj_tiles()

        if comps is None:
            comps = self.comps
        tod_shape = (self.fp.ndet, len(self.sight.Q))
        if dest is None:
            dest = np.zeros(tod_shape, np.float32)
        assert(dest.shape == tod_shape)  # P.fp/P.sight and dest argument disagree

        if self.tiled and self.pix_scheme == 'rectpix':
            # so3g <= 0.1.15 has a dims check on signal_map that fails on the tiled map format.
            so3g.proj.wcs._ProjectionistBase.from_map(
                proj, self._prepare_map(signal_map), self._get_asm(), signal=dest, comps=comps)
        else:
            proj.from_map(self._prepare_map(signal_map), self._get_asm(), signal=dest, comps=comps)

        if wrap is not None:
            if wrap in tod:
                del tod[wrap]
            tod.wrap(wrap, dest, [(0, 'dets'), (1, 'samps')])

        return dest

    @property
    def tiled(self):
        """Duck-typing to see if we're tiled or not. Reload-safe, unlike isinstance"""
        try:
            self.geom.ntile
            return True
        except AttributeError:
            return False

    def _comp_count(self, comps=None):
        """Returns the number of spin components for component code comps.

        """
        if comps is None:
            comps = self.comps
        return len(comps)

    def _get_proj(self):
        # Backwards compatibility for old so3g
        interpol_kw = _get_interpol_args(self.interpol)
        if self.geom is None:
            raise ValueError("Can't project without a geometry!")
        if self.pix_scheme == "healpix":
            return so3g.proj.ProjectionistHealpix.for_healpix(
                self.geom.nside, self.geom.nside_tile, self.active_tiles,
                self.geom.ordering, **interpol_kw)
        elif self.pix_scheme == "rectpix":
            if self.tiled:
                return so3g.proj.Projectionist.for_tiled(
                    self.geom.shape, self.geom.wcs, self.geom.tile_shape,
                    active_tiles=self.active_tiles, **interpol_kw)
            else:
                return so3g.proj.Projectionist.for_geom(self.geom.shape,
                    self.geom.wcs, **interpol_kw)

    def _get_proj_tiles(self, assign=False):
        # Get Projectionist and compute self.active_tiles if it's not
        # already known.  Return Projectionist with active_tiles set,
        # which is suitable for from_map and zeros (though not for
        # threaded to_map etc).
        proj = self._get_proj()
        if not self.tiled or (self.active_tiles is not None and not assign):
            return proj, {}
        tile_info = proj.get_active_tiles(self._get_asm(), assign=assign)
        self.active_tiles = tile_info['active_tiles']

        if self.pix_scheme == "healpix":
            self.geom.nside_tile = proj.nside_tile # Update nside_tile if it was 'auto'
            self.geom.ntile = 12*proj.nside_tile**2
        elif self.pix_scheme == 'rectpix':
            # Promote geometry to one with the active tiles marked.
            self.geom = tilemap.geometry(
                self.geom.shape, self.geom.wcs, self.geom.tile_shape,
                active=self.active_tiles)

        return self._get_proj(), tile_info

    def _get_proj_threads(self, cuts=None):
        """Return the Projectionist and sample-thread assignment for the
        present geometry.  If the thread assignment has not been
        determined yet, it is done now and cached in self.threads.  In
        tiled geometries, if self.active_tiles has not been
        determined, that is done now and cached.

        The returned sample-thread assignment is modified by "cuts",
        which defaults to self.cuts if passed as None.  I.e. after
        computing or looking up the full self.threads, the code
        returns (proj, self.threads*~cuts).

        Returns:
          Tuple (proj, threads*cuts).

        """
        proj = self._get_proj()
        if cuts is None:
            cuts = self.cuts

        if self.threads is None:
            if self.pix_scheme == "rectpix":
                self.threads = 'domdir'
            elif self.pix_scheme == "healpix":
                self.threads = 'tiles' if (self.geom.nside_tile is not None) else 'simple'

        need_tiles = (self.active_tiles is None)
        need_assign = (self.threads in ['tiles'])
        if need_tiles or need_assign:
            proj, tile_info = self._get_proj_tiles(need_assign)
        if need_assign:
            _tile_threads = wrap_ranges(tile_info['group_ranges'])

        if self.threads is False:
            return proj, ~cuts

        if isinstance(self.threads, str):
            if self.threads in ['simple', 'domdir']:
                logger.info(f'_get_proj_threads: assigning using "{self.threads}"')
                self.threads = wrap_ranges(proj.assign_threads(
                    self._get_asm(), method=self.threads))
            elif self.threads == 'tiles':
                # Computed above unless logic failed us...
                self.threads = _tile_threads
            else:
                raise ValueError('Request for unknown algo threads="%s"' % self.threads)
        if cuts:
            threads = [_t * ~cuts for _t in self.threads]
        else:
            threads = self.threads
        return proj, threads

    def _get_asm(self):
        """Bundles self.fp and self.sight into an "Assembly" for calling
        so3g.proj routines."""
        return so3g.proj.Assembly.attach(self.sight, self.fp)

    def _prepare_map(self, map):
        """Gently reformat a map in order to send it to so3g."""
        if self.tiled and self.pix_scheme == "rectpix":
            return list(map.tiles)
        else:
            return map

    def _flatten_map(self, map, uf_base=None):
        """Get a version of the map that is a numpy array, for passing
        to per-pixel math operations.  Relies on (self.pix_scheme,
        self.tiled) to interpret map.

        This also tries to extract wcs info (if rectpix) from the map,
        for inline consistency checking (e.g. so we're not happily
        projecting into a map we loaded from disk that has the same
        shape but is off by a few pixels from what pmat thinks is the
        right footprint).  It also looks at active_tiles / tile_list,
        and stores that for downstream compatibility checking with
        other flattened maps.

        If uf_base is passed in, it should be an unflatten_info dict
        (likely from a previous call to _flatten_map).  The analysis
        here will be checked against it for compatibility and any
        missing values (ahem wcs) will be used to augment the
        unflatten_info that is returned.

        Returns:
          array: The map, reformatted as an array (could simply be the
            input arg map, or a view of that, or a copy if necessary).
          unflatten_info: dict with misc compatibility info.

        """
        ufinfo = {'pix_scheme': self.pix_scheme,
                  'tiled': self.tiled}
        wcs = None
        crit_dims = 1
        if self.pix_scheme == 'healpix':
            if self.tiled:
                ufinfo['tile_list'] = [_m is not None for _m in map]
                map = hp_utils.tiled_to_compressed(map, -1)
                crit_dims = 2
            else:
                pass
        elif self.pix_scheme == 'rectpix':
            if self.tiled:
                if isinstance(map, tilemap.TileMap):
                    wcs = map.geometry.wcs
                    ufinfo['active_tiles'] = list(map.active)
                    ufinfo['tile_geom'] = map.geometry.copy(pre=())
            else:
                if isinstance(map, enmap.ndmap):
                    wcs = map.wcs
                crit_dims = 2
        ufinfo.update({'wcs': wcs,
                       'crit_dims': crit_dims,
                       'shape': map.shape})
        if uf_base is not None:
            ufinfo['wcs'] = _check_compat(uf_base, ufinfo)
        return map, ufinfo

    def _unflatten_map(self, map, uf_info):
        """Restore a map to full format, assuming it's currently an
        ndarray.  Intended as the inverse op to _flatten_map.
        Minimize the use of cached self.* here ... rely instead on
        uf_info.

        """
        if uf_info['pix_scheme'] == 'healpix':
            if uf_info['tiled']:
                map = hp_utils.compressed_to_tiled(map, uf_info['tile_list'], -1)
            else:
                pass
        elif uf_info['pix_scheme'] == 'rectpix':
            if uf_info['tiled']:
                if not isinstance(map, tilemap.TileMap):
                    g = uf_info['tile_geom']
                    g = tilemap.geometry(map.shape[:-1] + g.shape, g.wcs, g.tile_shape,
                                         active=uf_info['active_tiles'])
                    map = tilemap.TileMap(map, g)
            else:
                if not isinstance(map, enmap.ndmap) and uf_info['wcs']:
                    map = enmap.ndmap(map, uf_info['wcs'])
        return map


class P_PrecompDebug:
    def __init__(self, geom, pixels, phases):
        self.geom   = wrap_geom(geom).nopre
        self.pixels = pixels
        self.phases = phases
    def zeros(self, super_shape=None):
        if super_shape is None: super_shape = (self.phases.shape[2],)
        return enmap.zeros(super_shape + self.geom.shape, self.geom.wcs)
    def to_map(self, dest=None, signal=None, comps=None):
        if dest is None: dest  = self.zeros()
        proj = so3g.ProjEng_Precomp_NonTiled()
        proj.to_map(dest, self.pixels, self.phases, signal, None, None)
        return dest
    def from_map(self, signal_map, dest=None, comps=None):
        if dest is None: dest = np.zeros(self.pixels.shape[:2], np.float32)
        proj  = so3g.ProjEng_Precomp_NonTiled()
        proj.from_map(signal_map, self.pixels, self.phases, dest)
        return dest

def wrap_geom(geom):
    if isinstance(geom, tuple) or isinstance(geom, list):
        return enmap.Geometry(*geom)
    else:
        return geom

# Helpers for backwards compatibility with old so3g. Consider removing
# these once transition is done.  The idea is for sotodlib to use the
# more recent interface ("threads" is a list of RangesMatrix objects)
# while supporting use with older so3g (where "threads" is a single
# 3-d RangesMatrix).

def wrap_ranges(ranges):
    # Run this on the "threads" result returned from so3g thread
    # assignement routines, before storing the result internally.
    if _so3g_ivals_format() == 1:
        return [ranges]
    else:
        return ranges

def unwrap_ranges(ranges):
    # Run this on the internally stored threads object before passing
    # it to so3g projection routines.
    if _so3g_ivals_format() == 1:
        assert len(ranges) == 1, "Old so3g only supports simple (1-bunch) thread ranges, but got thread ranges with shape %s" % (str(ranges.shape))
        return ranges[0]
    else:
        return ranges

def _so3g_ivals_format():
    projclass = so3g.proj.Projectionist
    if not hasattr(projclass, '_ivals_format'):
        return 1
    else:
        return projclass._ivals_format

def _get_interpol_args(interpol):
    if _so3g_ivals_format() >= 2:
        return {'interpol': interpol}
    assert interpol in [None, "nn", "nearest"], "Old so3g does not support interpolated mapmaking"
    return {}

def _infer_pix_scheme(geom):
    fail_err = ValueError(f"Cannot determine pix_scheme from geom {geom}")

    # Healpix: geom contains 'nside' attribute
    if hasattr(geom, "nside"):
        pix_scheme = "healpix"

    # Rectpix: geom is None, an enmap Geometry or TileGeometry or a tuple/list that can be converted to one
    elif isinstance(geom, enmap.Geometry) or isinstance(geom, tilemap.TileGeometry):
        pix_scheme = "rectpix"
    elif isinstance(geom, tuple) or isinstance(geom, list):
        try:
            enmap.Geometry(*geom)
            pix_scheme = "rectpix"
        except:
            raise fail_err
    elif geom is None:
        pix_scheme = "rectpix"
    else:
        raise fail_err
    return pix_scheme

def _check_compat(*uf_infos):
    # Given one or more "uf_info" dicts, as returned by _flatten_map,
    # check that the flattened arrays are pixel-correspondent,
    # including (for rectpix cases) the wcs.
    #
    # Raises an error if any of that doesn't pan out.  On success,
    # returns the agreed-upon wcs (which could be None).
    ref_uf = uf_infos[0]
    for uf in uf_infos[1:]:
        for k in ['pix_scheme', 'tiled', 'active_tiles']:
            if ref_uf.get(k) != uf.get(k):
                raise ValueError(f"Inconsistent map structures: {uf_infos}")
        # Not sure how to handle broadcasting of lefter dims, so focus
        # on the pixel dim(s).
        dims_to_check = ref_uf['crit_dims']
        if ref_uf['shape'][-dims_to_check:] != uf['shape'][-dims_to_check:]:
            raise ValueError(f"Non-broadcastable map shapes: {uf_infos}")

    # And the wcs.
    wcss = [uf.get('wcs') for uf in uf_infos]
    wcs_to_use = _confirm_wcs(*wcss)

    return wcs_to_use



# Copyright (c) 2024 Simons Observatory.
# Full license can be found in the top level "LICENSE" file.

"""Check functionality of coords.pointing_model.

"""

import itertools
import unittest
import numpy as np

from sotodlib import coords, core
from so3g.proj import quat
from pixell import enmap

pm = coords.pointing_model

DEG = np.pi/180

def to_rad(*arg):
    return [a * DEG for a in arg]

def to_deg(*arg):
    return [a / DEG for a in arg]

def full_vectors(az=0., el=0., roll=0.):
    z = 0 * az + 0*el + 0*roll
    return z + az, z + el, z + roll

def center_branch(x, period=360., center=0):
    # Shift values in x by multiples of period, so they're within
    # period/w of center.
    branch = center - period / 2
    return (x - branch) % period + branch


class CoordsUtilsTest(unittest.TestCase):
    def test_populate(self):
        # Test that model populates tod.
        tod = core.AxisManager(core.OffsetAxis('samps', 1000))
        ancil = core.AxisManager(tod.samps)
        for k, v in zip(['az', 'el', 'boresight'],
                        full_vectors(np.linspace(0, 360, tod.samps.count), 60, 0)):
            ancil.wrap_new(f'{k}_enc', shape=('samps',))[:] = v
        tod.wrap('ancil', ancil)
        tod.wrap('obs_info', core.AxisManager())
        tod['obs_info'].wrap('telescope', 'satp1')

        results = []
        # TOD without 'pointing_model' -- should work, with warning
        with self.assertLogs(pm.__name__, level='WARN'):
            pm.apply_pointing_model(tod)
        results.append(tod.boresight)
        del tod['boresight']

        tod.wrap('pointing_model', core.AxisManager())
        tod['pointing_model'].wrap('version', 'sat_v1')
        pm.apply_pointing_model(tod, wrap=False)
        assert 'boresight' not in tod

        pm.apply_pointing_model(tod)
        assert 'boresight' in tod
        assert 'az' in tod.boresight
        results.append(tod.boresight)

        pm.apply_pointing_model(tod)

        # Check consistency...
        for r in results[1:]:
            for k in ['az', 'el', 'roll']:
                d = center_branch(r[k] - results[0][k], 2*np.pi)
                np.testing.assert_array_almost_equal(d, 0*d)

    def test_sat_v1(self):
        # Test model_sat_v1 general behaviors.
        az, el, roll = full_vectors(az=np.linspace(-90, 90, 100),
                                    el=60)
        params = core.AxisManager()
        params.wrap('version', 'sat_v1')
        az1, el1, roll1 = to_deg(*pm.model_sat_v1(params, *to_rad(az, el, roll)))
        params0 = {'version': 'sat_v1'}
        az1, el1, roll1 = to_deg(*pm.model_sat_v1(params, *to_rad(az, el, roll)))

        params = dict(params0)
        params['enc_offset_az'] = 1. * DEG
        az1, el1, roll1 = to_deg(*pm.model_sat_v1(params, *to_rad(az, el, roll)))
        assert np.all(center_branch(az1 - az) > 0)

        def tpoint_base_tilt(an, aw, az, el):
            c, s = np.cos(az * DEG), np.sin(az * DEG)
            delta_az = (-aw * c - an * s) * np.tan(el * DEG)
            delta_el = aw * s - an * c
            return az + delta_az / DEG, el + delta_el / DEG

        # Trial a few different base tilt combinations.
        az, el, roll = full_vectors(az=np.arange(0, 360), el=30.)
        tpoint_err_max = 0.01 # deg
        for (bt_c, bt_s, high_az) in [
                (1 * DEG, 0 * DEG, 0.),
                (0 * DEG, 1 * DEG, 270.),
                (2**-0.5 * DEG, 2**-0.5 * DEG, 315.),
        ]:
            params = dict(params0)
            params.update({
                'base_tilt_cos': bt_c,
                'base_tilt_sin': bt_s,  # West up.
                })
            az1, el1, roll1 = to_deg(*pm.model_sat_v1(params, *to_rad(az, el, roll)))

            # Check that delta el, near the expected high_az point, is postive.
            d_el = el1 - el
            s = abs(center_branch(az - high_az)) < 10.
            assert (all(d_el[s]) > 0)

            # Check consistency with tpoint linear approximation --
            # bt_c and bt_s correspond to AN and AW terms,
            # respectively.
            az2, el2 = tpoint_base_tilt(bt_c, bt_s, az, el)
            d_az = abs(center_branch(az2 - az1))
            d_el = abs(center_branch(el2 - el1))
            assert all(d_az < tpoint_err_max)
            assert all(d_el < tpoint_err_max)

        # Boresight center.
        params = dict(params0)
        params.update({
            'fp_rot_xi0': 0.4 * DEG,
            'fp_rot_eta0': 0.1 * DEG,
            #'fp_offset_xi0': 0.05 * DEG,
            #'fp_offset_eta0': 0.01 * DEG,        
            })
        az, el, roll = full_vectors(roll=np.linspace(0., 360., 20), el=45.)
        for xi, eta, sig in [
                (0.4, 0.1, 0.),
                (0.5, 0.2, 0.1*2**.5),
        ]:
            az1, el1, roll1 = pm.model_sat_v1(params, *to_rad(az, el, roll))
            # Measure az, el on sky of a detector (xi, eta):
            q_hs1 = (quat.rotation_lonlat(-az1, el1, roll1) *
                     quat.rotation_xieta(xi * DEG, eta * DEG))
            neg_az2, el2, roll2 = quat.decompose_lonlat(q_hs1)
            d_el = el2 - el * DEG
            d_azc = center_branch(-neg_az2 - az*DEG, 2*np.pi) * np.cos(el*DEG)
            sig_meas = (d_el.std()**2 + d_azc.std()**2)**.5
            assert(abs(sig_meas - sig * DEG) < .001*DEG)


if __name__ == '__main__':
    unittest.main()



import warnings
from dataclasses import dataclass, fields, asdict, field
from typing import List, Optional, Tuple, Iterator, Union
from copy import deepcopy
import h5py

import numpy as np
import matplotlib.pyplot as plt
from scipy.optimize import linear_sum_assignment

from sotodlib.core import AxisManager
from sotodlib.coords import optics
from sotodlib.core import metadata
from sotodlib.io.metadata import write_dataset



def map_band_chans(b1, c1, b2, c2, chans_per_band=512):
    """
    Returns an index mapping of length nchans1 from one set of bands and
    channels to another. Note that unmapped indices are returned as -1, so this
    must be handled before (or after) indexing or else you'll get weird
    results.
    Args:
        b1 (np.ndarray):
            Array of length nchans1 containing the smurf band of each channel
        c1 (np.ndarray):
            Array of length nchans1 containing the smurf channel of each channel
        b2 (np.ndarray):
            Array of length nchans2 containing the smurf band of each channel
        c2 (np.ndarray):
            Array of length nchans2 containing the smurf channel of each channel
        chans_per_band (int):
            Lets just hope this never changes.
    """
    acs1 = b1 * chans_per_band + c1
    acs2 = b2 * chans_per_band + c2

    mapping = np.full_like(acs1, -1, dtype=int)
    for i, ac in enumerate(acs1):
        x = np.where(acs2 == ac)[0]
        if len(x > 0):
            mapping[i] = x[0]
    return mapping


def get_north_is_highband(bands, bgs):
    """
    Checks if north is highband based on bgmapping. This will tell you if
    the majority of dets on the north side of the ufm (bgs 0-5) belong to
    highband (bands 4-7).
    """
    highband = (bands > 3).astype(int) * 2 - 1   # 1 if highband, -1 if not
    north = (bgs < 6).astype(int) * 2 - 1        # 1 if north, -1 if not
    north[bgs == -1] = 0
    return np.mean(highband * north) > 0


@dataclass
class PointingConfig:
    """
    Helper class for getting pointing info from an optics model.

    Args
    -----
    fp_file: str
        Path to focal-plane file that is used by the optics module.
    wafer_slot: str
        Wafer slot of the UFM. For example: "ws0"
    tel_type: str
        Tel type for the optics model. Either "SAT" or "LAT"
    zemax_path: str
        If running for a "LAT" tel_type, the path to the zemax file must be specified.
    roll: float
        Rotation about the line of sight.
        For the LAT this is elev - 60 - corotator.
        For the SAT this is -1*boresight.
    tube_slot: str/int
        If running for a "LAT" tel_type, the tube slot must be specified.
        Either the tube name as a string or the tube number as an int.
    """
    fp_file: str
    wafer_slot: str
    tel_type: str
    zemax_path: Optional[str] = None
    roll: Optional[float] = 0
    tube_slot: Optional[Union[str, int]] = None

    dx: float = field(init=False)
    dy: float = field(init=False)
    theta: float = field(init=False)
    fp_pars: dict = field(init=False)

    def __post_init__(self):
        if self.tel_type == 'LAT':
            if self.zemax_path is None:
                raise ValueError("zemax path must be set for 'LAT' tel_type")
            if self.tube_slot is None:
                raise ValueError("tube slot must be set for 'LAT' tel_type")

        if self.tel_type not in ['SAT', 'LAT']:
            raise ValueError("tel_type should be 'SAT' or 'LAT' ")

        self.fp_pars = optics.get_ufm_to_fp_pars(
            self.tel_type, self.wafer_slot, self.fp_file
        )
        self.dx = self.fp_pars['dx']
        self.dy = self.fp_pars['dy']
        self.theta = np.deg2rad(self.fp_pars['theta'])

    def get_pointing(self, x, y, pol=0):
        xp = x * np.cos(self.theta) - y * np.sin(self.theta) + self.dx
        yp = x * np.sin(self.theta) + y * np.cos(self.theta) + self.dy

        if self.tel_type.upper() == 'SAT':
            xi, eta, gamma = optics.SAT_focal_plane(
                None, x=xp, y=yp, pol=pol, roll=self.roll
            )
        elif self.tel_type.upper() == 'LAT':
            xi, eta, gamma = optics.LAT_focal_plane(
                None, self.zemax_path, x=xp, y=yp, pol=pol,
                roll=self.roll, tube_slot=self.tube_slot
            )
        return xi, eta, gamma


@dataclass
class Resonator:
    """
    Data structure to hold any resonator information.
    """
    idx: int

    is_north: int
    res_freq: float
    res_qi: float = np.nan
    smurf_res_idx: int = -1
    smurf_band: int = -1
    smurf_channel: int = -1
    smurf_subband: int = -1
    readout_id: str = ''

    xi: float = np.nan
    eta: float = np.nan
    gamma: float = np.nan

    bg: int = -1
    det_x: float = np.nan
    det_y: float = np.nan
    det_row: int = 0
    det_col: int = 0
    pixel_num: int = 0
    det_rhomb: str = ''
    det_pol: str = ''
    det_freq: int = 0
    det_bandpass: str = ''
    det_angle_raw_deg: float = np.nan
    det_angle_actual_deg: float = np.nan
    det_type: str = ''
    det_id: str = 'NO_MATCH'
    is_optical: int = 1
    mux_bondpad: int = 0
    mux_subband: str = ''
    mux_band: int = -1
    mux_channel: int = -1
    mux_layout_pos: int = -1

    matched: int = 0
    match_idx: int = -1


def apply_design_properties(smurf_res, design_res, in_place=False, apply_pointing=True):
    """
    Combines two resonators into one, taking smurf-properties such as res-idx,
    smurf-band, and smurf-channel one, and design properties such as det
    position and polarization from the other.

    Args:
        smurf_res (Resonator):
            The resonator to take smurf properties from
        design_res (Resonator):
            The resonator to take design properties from
        in_place (bool):
            If True, the src_res will be modified. Otherwise, a new Resonator
            will be created and returned.
    """
    if in_place:
        r = smurf_res
    else:
        r = deepcopy(smurf_res)

    design_props = [
        'bg', 'det_x', 'det_y', 'det_row', 'det_col', 'pixel_num', 'det_rhomb',
        'det_pol', 'det_freq', 'det_bandpass', 'det_angle_raw_deg',
        'det_angle_actual_deg', 'det_type', 'det_id', 'is_optical',
        'mux_bondpad', 'mux_subband', 'mux_band', 'mux_channel',
        'mux_layout_pos'
    ]
    if apply_pointing:
        design_props += ['xi', 'eta', 'gamma']

    for prop in design_props:
        setattr(r, prop, getattr(design_res, prop))

    return r


class ResSet:
    """
    Class to hold a group of resonances. This provides easy interfaces for
    accessing Resonance fields as np arrays for fast computations
    and provides initialization functions from different data sources.
    """

    def __init__(self, resonances: List[Resonator], name=None):
        self.resonances: List[Resonator] = resonances
        self.name = name

    def __iter__(self):
        yield from self.resonances

    def __getitem__(self, k):
        return self.resonances[k]

    def __len__(self):
        return len(self.resonances)

    @classmethod
    def from_array(cls, arr, name=None, ignore_extra_fields=True):
        """
        Creates a ResSet from a numpy structured array (resulting from
        ``as_array`` method). 

        Args
        -----
        arr: np.ndarray
            Structured ResSet array
        name: str
            Name for the res-set
        ignore_extra_fields: bool
            If True, this will ignore any fields from the array that are not in
            the Resonator dataclass. This may happen if loading an older saved array,
            where the resonator fields are not the same.
        """
        resonators = []
        names = arr.dtype.names
        field_names = [f.name for f in fields(Resonator)]
        for a in arr:
            # Remove any fields that are not in the Resonator class if
            # ignore_extra_fields is True.
            kw = dict(zip(names, a))
            for name in names:
                if name in field_names:
                    continue
                if ignore_extra_fields:
                    del kw[name]
                else:
                    raise ValueError(f"Field '{name}' not in Resonator class.")
            resonators.append(Resonator(**kw))
        return cls(resonators, name=name)


    def as_array(self):
        """
        Returns resonance data in the form of a numpy structured array.
        """
        dtype = []
        data = []
        for field in fields(Resonator):
            if field.type == str:
                typ = '<U50'
            else:
                typ = field.type
            dtype.append((field.name, typ))
        for r in self.resonances:
            data.append(tuple(getattr(r, name) for name, _ in dtype))
        return np.array(data, dtype=dtype)

    @classmethod
    def from_aman(cls, aman, stream_id, det_cal=None, name=None, pointing: Optional[AxisManager]=None):
        """
        Load a resonator set from a Context object based on an obs_id

        Args
        ----------
        aman: AxisManager
            Axis manager containing metadata
        stream_id: str
            Stream id for ResSet to load
        det_cal: AxisManager
            Detector calibration metadata. If not specified, will default to
            ``aman.det_cal``
        pointing: Optional[AxisManager]
            AxisManager containing pointing metadata. If set, this should be an
            AxisManager containing the fields ``xi`` and ``eta``, and resonator
            pointing information will be added from here.
        """
        m = aman.det_info.stream_id == stream_id
        if not np.any(m):
            raise ValueError(f"No channels with stream_id {stream_id} in obs")
        
        if det_cal is None:
            det_cal = aman.det_cal
        north_is_highband = get_north_is_highband(
            aman.det_info.smurf.band[m], det_cal.bg[m]
        )
        resonators = []
        for i, ri in enumerate(np.where(m)[0]):
            band, channel = aman.det_info.smurf.band[ri], aman.det_info.smurf.channel[ri]
            is_north = north_is_highband ^ (band < 4)
            readout_id = aman.det_info.readout_id[ri]
            bg = det_cal.bg[ri]
            res_freq=aman.det_info.smurf.frequency[ri]
            if res_freq >= 6000:
                res_freq -= 2000

            res = Resonator(
                idx=i, is_north=is_north, res_freq=res_freq, smurf_band=band,
                smurf_channel=channel, readout_id=readout_id, bg=bg
            )
            if pointing is not None:
                res.xi = pointing.xi[ri]
                res.eta = pointing.eta[ri]

            resonators.append(res)

        return cls(resonators, name=name)

    @classmethod
    def from_tunefile(cls, tunefile, name=None, north_is_highband=True,
                      resfit_file=None, bgmap_file=None):
        """
        Creates an instance based on a smurf-tune file. If a resfit or bgmap
        file is included, that data will be added to the Resonance objects as
        well.

        Args:
            tunefile (str):
                Path to Pysmurf tunefile
            name (str):
                Name to label this ResSet
            north_is_highband (bool):
                True if the north-side of the array corresponds to bands
                4-7
            resfit_file (str):
                Path to file containing resonance fit data
            bgmap_file (str):
                Path to file containing bgmap data
        """
        tune = np.load(tunefile, allow_pickle=True).item()
        resonances = []
        idx = 0
        for band, _v in tune.items():
            if 'resonances' not in _v:
                continue

            for res_idx, d in _v['resonances'].items():
                is_north = north_is_highband ^ (band < 4)
                res = Resonator(
                    idx=idx, smurf_res_idx=res_idx, res_freq=d['freq'],
                    smurf_band=band, is_north=is_north
                )
                if d['channel'] != -1:
                    res.smurf_channel = d['channel']

                if res.smurf_band >= 4:
                    res.res_freq -= 2000

                resonances.append(res)
                idx += 1
        rs = cls(resonances, name=name)
        if resfit_file is not None:
            rs.add_resfit_data(resfit_file)
        if bgmap_file is not None:
            rs.add_bgmap_data(bgmap_file)
        return rs

    @classmethod
    def from_wafer_info_file(cls, wafer_info_file, array_name, name=None,
                             pt_cfg: Optional[PointingConfig]=None):
        """
        Initialize a ResSet from a wafer info file. This is a file that contains
        detector design information.

        Args
        -----
        wafer_info_file: str
            Path to wafer info file
        array_name: str
            Array name, which is the key in the wafer-info-file. For example:
            "mv7".
        name: str
            Name to assign to the ResSet
        pt_cfg: PointingConfig
            If set, this will be used to get pointing info based on the optics
            model. If not set, pointing info will not be included in the ResSet.
        """

        with h5py.File(wafer_info_file) as f:
            wafer_array = np.array(f[array_name])

        resonators = []
        idx = 0
        for r in wafer_array:
            is_north = r['dets:wafer.coax'] == b'N'
            res = Resonator(
                idx=idx,
                det_id=r['dets:det_id'].decode(),
                mux_bondpad=r['dets:wafer.bond_pad'],
                mux_band=r['dets:wafer.mux_band'],
                mux_channel=r['dets:wafer.mux_channel'],
                mux_subband=r['dets:wafer.mux_subband'],
                mux_layout_pos=r['dets:wafer.mux_position'],
                res_freq=r['dets:wafer.design_freq_mhz'],
                bg=r['dets:wafer.bias_line'],
                det_pol=r['dets:wafer.pol'],
                det_bandpass=r['dets:wafer.bandpass'],
                det_row=r['dets:wafer.det_row'],
                det_col=r['dets:wafer.det_col'],
                det_rhomb=r['dets:wafer.rhombus'],
                det_type=r['dets:wafer.type'],
                det_x=r['dets:wafer.x'],
                det_y=r['dets:wafer.y'],
                det_angle_actual_deg=r['dets:wafer.angle'],
                is_north=is_north
            )

            if pt_cfg is not None:
                res.xi, res.eta, res.gamma = pt_cfg.get_pointing(res.det_x, res.det_y)

            resonators.append(res)
            idx += 1

        return cls(resonators, name=name)

    @classmethod
    def from_solutions(cls, sol_file, north_is_highband=True, name=None,
                       fp_pars=None, platform='SAT', zemax_path=None):
        """
        Creates an instance from an input-solution file. This will include both design data, along with smurf-band
        and smurf-channel info. Resonance frequencies used here are the VNA
        freqs measured by Kaiwen.

        Args:
            sol_file (str):
                Path to solutions file
            name (str):
                Name to label this ResSet
            north_is_highband (bool):
                True if the north-side of the array corresponds to bands
                4-8
            fp_pars (dict):
                Result of the function ``sotododlib.coords.optics.get_ufm_to_fp_pars``. If this is None, detector positions will
                not be mapped to pointing angles.
            platform (str):
                'SAT' or 'LAT'. Used to determine which focal plane function to
                use for pointing
            zemax_path (str):
                zemax path, required to get pointing for LAT optics
        """
        resonances = []

        if fp_pars is not None:
            theta = np.deg2rad(fp_pars['theta'])
            dx, dy = fp_pars['dx'], fp_pars['dy']

        with open(sol_file, 'r') as f:
            labels  = f.readline().split(',') # Read header
            for i in range(len(labels)):
                labels[i] = labels[i].strip()

            #Helper needed for when sol file has saved ints as floats
            def _int(val, null_val=None): 
                is_null = (val == 'null' or val == '')
                if is_null and null_val is not None:
                    return null_val
                return int(float(val))

            i = 0
            for line in f.readlines():
                d = dict(zip(labels, line.split(',')))
                if not d['bias_line']:  # Skip incomplete lines
                    continue
                # is_north = north_is_highband ^ (_int(d['smurf_band']) < 4)
                # is_north = d['is_north'].lower().strip() == 'true'
                is_north = _int(d['bias_line']) < 6
                is_optical = (d['is_optical'].lower() == 'true')
                r = Resonator(
                    i, res_freq=float(d['freq_mhz']), smurf_band=_int(d['smurf_band']),
                    bg=_int(d['bias_line']), det_x=float(d['det_x']),
                    det_y=float(d['det_y']), det_rhomb=d['rhomb'],
                    det_row=_int(d['det_row']), det_col=_int(d['det_col']),
                    pixel_num=_int(d['pixel_num'], null_val=-1),
                    det_type=d['det_type'], det_id=d['detector_id'].strip(),
                    mux_band=_int(d['mux_band']), mux_channel=_int(d['mux_channel']),
                    mux_subband=d['mux_subband'], mux_bondpad=d['bond_pad'],
                    det_angle_raw_deg=float(d['angle_raw_deg']),
                    det_angle_actual_deg=float(d['angle_actual_deg']),
                    mux_layout_pos=_int(d['mux_layout_position']),
                    det_bandpass=d['bandpass'], det_pol=d['pol'],
                    is_optical=is_optical,
                    is_north=is_north
                )

                if _int(d['smurf_channel']) != -1:
                    r.smurf_channel = _int(d['smurf_channel'])

                if fp_pars is not None:
                    x, y = float(d['det_x']), float(d['det_y'])
                    xp = x * np.cos(theta) - y * np.sin(theta) + dx
                    yp = x * np.sin(theta) + y * np.cos(theta) + dy
                    if platform.upper() == 'SAT':
                        xi, eta, gamma = optics.SAT_focal_plane(
                            None, x=xp, y=yp, pol=0
                        )
                    elif platform.upper() == 'LAT':
                        xi, eta, gamma = optics.LAT_focal_plane(
                            None, zemax_path, x=xp, y=yp, pol=0,
                        )
                    else:
                        raise ValueError(
                            f"Unknown platform {platform}. Must be 'SAT' or 'LAT'"
                        )
                    r.xi = xi
                    r.eta = eta
                    r.gamma = gamma

                resonances.append(r)
                i += 1

        return cls(resonances, name=name)

    def add_resfit_data(self, resfit_file):
        """
        Adds resonator quality data from a res_fit file.
        """
        resfits = np.load(resfit_file, allow_pickle=True).item()
        for r in self.resonances:
            d = resfits[r.smurf_band][r.smurf_res_idx]
            r.res_qi = d['derived_params']['Qi']

    def add_bgmap_data(self, bgmap_file):
        """
        Adds bias group data from an sodetlib bgmap file.
        """
        bgmap = np.load(bgmap_file, allow_pickle=True).item()
        bands = np.array([res.smurf_band for res in self.resonances], dtype=int)
        chans = np.array([res.smurf_channel for res in self.resonances], dtype=int)
        idxmap = map_band_chans(bands, chans, bgmap['bands'], bgmap['channels'])

        for i, r in enumerate(self.resonances):
            idx = idxmap[i]
            if idx == -1:
                continue
            bg = bgmap['bgmap'][idx]
            if bg != -1:
                r.bg = bg

    def add_pointing(self, bands, chans, xis, etas):
        """
        Adds measured detector pointing to resonators.
        """
        bs = np.array([res.smurf_band for res in self.resonances], dtype=int)
        cs = np.array([res.smurf_channel for res in self.resonances], dtype=int)
        idxmap = map_band_chans(bs, cs, bands, chans)
        for i, r in enumerate(self.resonances):
            idx = idxmap[i]
            if idx == -1:
                continue
            r.xi = xis[idx]
            r.eta = etas[idx]


@dataclass
class MatchParams:
    """
    Any constants / hardcoded values go here to be fiddled with

    Args:
        unassigned_slots (int):
            Number of additional "unassigned" node to use per-side
        freq_offset_mhz (float):
            constant offset between resonator-set frequencies to use in
            matching.
        freq_width (float):
            width of exponential to use in the frequency cost function (MHz).
        dist_width (float)
            width of exponential to use in the pointing cost function (rad)
        unmatched_good_res_pen (float):
            penalty to apply to leaving a resonator with a good qi unassigned
        good_res_qi_thresh (float):
            qi threshold that is considered "good"
        enforce_pointing_reqs (bool):
            If this is enabled, it will enforce the following requirements when
            matching:

             - Resonators with OPTC det_type must have pointing data.
             - Resonators with UNRT, SQID, or BARE det_types must _not_ have pointing data.
             - Resonators with DARK or SLOT det_types may or may not have pointing data.
        assigned_bg_unmatched_pen (float):
            Penalty to apply to leaving a resonator with an assigned bg
            unmatched
        unassigned_bg_unmatched_pen (float):
            Penalty to apply to leaving a resonator with an unassigned bg
            unmatched
        assigned_bg_mismatch_pen (float):
            Penalty to apply for matching a resonator with an assigned bias line
            to another one with a mis-matched bias line.
        unassigned_bg_mismatch_pen (float):
            Penalty to apply for matching a resonator with no bias line to
            another one with a mis-matched bias line.
    """
    unassigned_slots: int = 1000
    freq_offset_mhz: float = 0.0
    freq_width: float = 2.
    dist_width: float =0.01
    unmatched_good_res_pen: float = 10.
    good_res_qi_thresh: float = 100e3
    enforce_pointing_reqs: bool = False

    assigned_bg_unmatched_pen: float = 100000
    unassigned_bg_unmatched_pen: float = 10000
    assigned_bg_mismatch_pen: float = 100000
    unassigned_bg_mismatch_pen: float = 1


@dataclass
class MatchingStats:
    unmatched_src: int = 0
    unmatched_dst: int = 0
    unmatched_src_with_pointing: int = 0
    matched_chans: int = 0
    mismatched_bg: int = 0
    freq_diff_avg: float = 0.0
    freq_err_avg: float = 0.0
    pointing_err_avg: float = 0.0


class Match:
    """
    Class for performing a Resonance Matching between two sets of resonators,
    labeled `src` and `dst`. In the matching algorithm there is basically no
    difference between `src` and `dst` res-sets, except:

     - When merged, smurf-data such as band, channel, and res-idx will be taken
       from the ``src`` res-set, while detector information will be taken from
       the ``dst`` set.

    Args:
        src (ResSet):
            The source resonator set
        dst (ResSet):
            The dest resonator set
        match_pars (MatchParams):
            MatchParams object used in the matching algorithm. This
            can be used to tune the cost-function and matching.
        apply_dst_pointing (bool):
            If True, the ``merged`` res-set will take its pointing information
            from ``dst`` instead of ``src``.

    Attributes:
        src (ResSet):
            The source resonator set
        dst (ResSet):
            The dest resonator set
        match_pars (MatchParams):
            MatchParams object used in the matching algorithm. This
            can be used to tune the cost-function and matching.
        matching (np.ndarray):
            A 2xN array of indices, where the first row corresponds to the
            indices of the src resonators, and the second row corresponds to
            the indices of the dst resonators.  If the index is larger than the
            size of the corresponding res-set, that means the paired
            resonator was not matched.
        merged (ResSet):
            A ResSet containing the merged resonators. This is created by
            applying the "design" properties of the dst resonators to the source
            resonators. If a source resonator is not matched to any dest
            resonator, it will be copied as-is.
        stats (MatchingStats):
            A MatchingStats object containing some statistics about the
            matching.
    """
    def __init__(self, src: ResSet, dst: ResSet, match_pars: Optional[MatchParams]=None,
                 apply_dst_pointing=True):
        self.src = src
        self.dst = dst

        if match_pars is None:
            self.match_pars = MatchParams()
        else:
            self.match_pars = match_pars
        self.apply_dst_pointing = apply_dst_pointing

        self.matching_cost = np.nan
        self.matching, self.merged = self._match()
        self.stats = self.get_stats()

    def _get_biadjacency_matrix(self) -> np.ndarray:
        src_arr = self.src.as_array()
        dst_arr = self.dst.as_array()

        mat = np.zeros((len(self.src), len(self.dst)), dtype=float)

        # N/S mismatch
        m = src_arr['is_north'][:, None] != dst_arr['is_north'][None, :]
        mat[m] = np.inf

        if self.match_pars.enforce_pointing_reqs:
            # Det types of DARK and SLOT are allowed to may or may not have
            # pointing data. For these types of detectors, the cost is left
            # untouched.

            src_has_pointing = np.isfinite(src_arr['xi']) & np.isfinite(src_arr['eta'])
            dst_has_pointing = np.isfinite(dst_arr['xi']) & np.isfinite(dst_arr['eta'])

            src_no_match = np.isin(src_arr['det_type'], ['NC'])
            dst_no_match = np.isin(dst_arr['det_type'], ['NC'])
            mat[src_no_match, :] = np.inf
            mat[:, dst_no_match] = np.inf

            src_pointing_forbidden = np.isin(src_arr['det_type'], ['UNRT', 'SQID', 'BARE'])
            dst_pointing_forbidden = np.isin(dst_arr['det_type'], ['UNRT', 'SQID', 'BARE'])
            m = src_pointing_forbidden[:, None] & dst_has_pointing[None, :]
            mat[m] = np.inf
            m = src_has_pointing[:, None] & dst_pointing_forbidden[None, :]
            mat[m] = np.inf

            src_pointing_required = np.isin(src_arr['det_type'], ['OPTC'])
            dst_pointing_required = np.isin(dst_arr['det_type'], ['OPTC'])
            m = src_pointing_required[:, None] & (~dst_has_pointing[None, :])
            mat[m] = np.inf
            m = (~src_has_pointing[:, None]) & dst_pointing_required[None, :]
            mat[m] = np.inf

        # Frequency offset
        df = src_arr['res_freq'][:, None] - dst_arr['res_freq'][None, :]
        df -= self.match_pars.freq_offset_mhz
        mat += np.exp((np.abs(df / self.match_pars.freq_width)) ** 2)

        # BG mismatch
        bgs_mismatch = src_arr['bg'][:, None] != dst_arr['bg'][None, :]
        bgs_unassigned = (src_arr['bg'][:, None] == 1) | (dst_arr['bg'][None, :] == -1)

        m = bgs_mismatch & bgs_unassigned
        mat[m] += self.match_pars.unassigned_bg_mismatch_pen
        m = bgs_mismatch & (~bgs_unassigned)
        mat[m] += self.match_pars.assigned_bg_mismatch_pen

        # If pointing, add cost if assigned too far
        dd = np.sqrt(
              (src_arr['xi'][:, None] - dst_arr['xi'][None, :])**2 \
            + (src_arr['eta'][:, None] - dst_arr['eta'][None, :])**2)
        m = ~np.isnan(dd)
        mat[m] += np.exp((np.abs(dd[m] / self.match_pars.dist_width)) ** 2)

        # Any remaining nans should be set to info so matrix is still solvable
        mat[np.isnan(mat)] = np.inf

        return mat

    def _get_unassigned_costs(self, rs):
        ra = rs.as_array()

        arr = np.zeros(len(rs))

        # Additional cost for leaving good resonator unassigned
        is_good_res = ra['res_qi'] > self.match_pars.good_res_qi_thresh
        is_good_res[np.isnan(is_good_res)] = 0
        arr[is_good_res] += self.match_pars.unmatched_good_res_pen

        bg_assigned = ra['bg'] != -1
        arr[bg_assigned] += self.match_pars.assigned_bg_unmatched_pen
        arr[~bg_assigned] += self.match_pars.unassigned_bg_unmatched_pen

        return arr


    def _match(self):
        nside = max(len(self.src), len(self.dst)) + self.match_pars.unassigned_slots

        # Keep this square so all resonators are included in final matching
        mat_full = np.zeros((nside, nside), dtype=float)
        with warnings.catch_warnings():
            warnings.filterwarnings(
                action='ignore',
                message='overflow encountered in exp*',
                category=RuntimeWarning
            )

            mat_full[:len(self.src), :len(self.dst)] = self._get_biadjacency_matrix()
            mat_full[:len(self.src), len(self.dst):] = \
                self._get_unassigned_costs(self.src)[:, None]
            mat_full[len(self.src):, :len(self.dst)] = \
                self._get_unassigned_costs(self.dst)[None, :]
            mat_full[len(self.src):, len(self.dst):] = 0

        self.matching = np.array(linear_sum_assignment(mat_full))
        self.matching_cost = mat_full[self.matching[0], self.matching[1]].sum()

        for r1, r2 in self.get_match_iter(include_unmatched=True):
            if r1 is None:
                r2.matched = 0
                continue
            if r2 is None:
                r1.matched = 0
                continue

            r1.matched = 1
            r1.match_idx = r2.idx
            r2.matched = 1
            r2.match_idx = r1.idx

        resonances = []
        for r1 in self.src:
            if r1.matched:
                r2 = self.dst[r1.match_idx]
                r = apply_design_properties(
                    r1, r2, in_place=False,
                    apply_pointing=self.apply_dst_pointing
                )
            else:
                r = deepcopy(r1)
            resonances.append(r)
        self.merged = ResSet(resonances, name='merged')

        return self.matching, self.merged

    def get_match_iter(self, include_unmatched=True) \
        -> Iterator[Tuple[Optional[Resonator], Optional[Resonator]]]:
        """
        Returns an iterator over matched resonators (r1, r2).

        Args:
            include_unmatched (bool):
                If True, will include unmatched resonators, with the pair set
                to None.
        """
        for i, j in zip(*self.matching):
            if i < len(self.src):
                r1 = self.src[i]
            else:
                r1 = None
            if j < len(self.dst):
                r2 = self.dst[j]
            else:
                r2 = None

            if (r1 is None) and (r2 is None):
                continue

            if (r1 is None) or (r2 is None):
                if not include_unmatched:
                    continue

            yield r1, r2

    def get_stats(self) -> MatchingStats:
        """
        Gets stats associated with current matching.
        """
        stats = MatchingStats()
        src_arr = self.src.as_array()
        dst_arr = self.dst.as_array()

        stats.unmatched_src = np.sum(~src_arr['matched'].astype(bool))
        stats.unmatched_dst = np.sum(~dst_arr['matched'].astype(bool))

        has_pt = ~np.isnan(src_arr['xi'])
        stats.unmatched_src_with_pointing = np.sum(
            (~src_arr['matched'].astype(bool)) & has_pt
        )

        dangs = []
        for r1, r2 in self.get_match_iter(include_unmatched=False):
            stats.matched_chans += 1
            if r1.bg != r2.bg:
                stats.mismatched_bg += 1
            stats.freq_diff_avg += r1.res_freq - r2.res_freq
            stats.freq_err_avg += np.abs(
                r1.res_freq - r2.res_freq - self.match_pars.freq_offset_mhz
            )
            dangs.append(
                np.sqrt((r1.xi - r2.xi)**2 + (r1.eta - r2.eta)**2)
            )

        stats.pointing_err_avg = np.nan
        if (~np.isnan(dangs)).any():
            stats.pointing_err_avg = np.nanmean(dangs)

        stats.freq_diff_avg /= stats.matched_chans
        stats.freq_diff_avg = float(stats.freq_diff_avg)

        stats.freq_err_avg /= stats.matched_chans
        stats.freq_err_avg = float(stats.freq_err_avg)

        return stats

    def save(self, path):
        """Saves match to HDF5 file"""
        with h5py.File(path, 'w') as fout:
            fout.create_group('meta')
            fout.create_group('meta/match_pars')
            for k, v in asdict(self.match_pars).items():
                fout['meta/match_pars'][k] = v
            if self.src.name is not None:
                fout['meta/src_name'] = self.src.name
            if self.dst.name is not None:
                fout['meta/dst_name'] = self.dst.name

            write_dataset(
                metadata.ResultSet.from_friend(self.src.as_array()), fout, 'src')
            write_dataset(
                metadata.ResultSet.from_friend(self.dst.as_array()), fout, 'dst')
            write_dataset(np.array(self.matching), fout, 'matching')
            write_dataset(
                metadata.ResultSet.from_friend(self.merged.as_array()), fout, 'merged')

    @classmethod
    def load(cls, path):
        """
        Loads a match from a h5 file (resulting from ``save`` method)
        """
        with h5py.File(path) as f:
            src = ResSet.from_array(np.array(f['src']))
            if 'meta/src_name' in f:
                src.name = f['meta/src_name'][()].decode()
            dst = ResSet.from_array(np.array(f['dst']))
            if 'meta/dst_name' in f:
                dst.name = f['meta/dst_name'][()].decode()
            match_pars = {}
            match_par_fields = [f.name for f in fields(MatchParams)]
            for k in f['meta/match_pars'].keys():
                if k not in match_par_fields:
                    print(f"Ignoring unknonwn kwarg for match-param: {k}")
                else:
                    match_pars[k] = f['meta/match_pars'][k][()]
            match_pars = MatchParams(**match_pars)
        match = cls(src, dst, match_pars=match_pars)
        return match


def plot_match_freqs(m: Match, is_north=True, show_offset=False, xlim=None):
    """
    Plots src and dst ResSet freqs in a match, along with their matching
    assignments.
    """

    fig, ax = plt.subplots(figsize=(10, 6))

    if not show_offset:
        offset = m.match_pars.freq_offset_mhz
    else:
        offset = 0

    yw= 0.4
    def plot_res(ax, r: Resonator, y=0, offset=0):
        if r is None:
            return
        if r.is_north != is_north:
            return

        xs = [r.res_freq+offset, r.res_freq+offset]
        ys = [y - yw, y+yw]

        if r.bg == -1:
            c = 'grey'
        else:
            c = f'C{r.bg}'

        ax.plot(xs, ys, c=c)

    for r1, r2 in m.get_match_iter():
        plot_res(ax, r1, y=0)
        plot_res(ax, r2, y=1, offset=offset)

    for r1, r2 in m.get_match_iter(include_unmatched=False):
        if r1.is_north != is_north:
            continue
        ax.plot([r1.res_freq, r2.res_freq+offset], [yw, 1-yw], 
                c='grey', alpha=.3)

    ax.set_yticks([yw/2, 1-yw/2])
    labels = ['ResSet1', 'ResSet2']
    for i, rs in enumerate([m.src, m.dst]):
        if rs.name is not None:
            labels[i] = rs.name

    ax.set_yticklabels(labels, fontsize=14, rotation=90, va='center')

    ax.set_ylim(0, 1)
    if xlim is not None:
        ax.set_xlim(*xlim)

    ax.set_xlabel("Resonance Freq (MHz)", fontsize=14)
    ax.set_title(f"Freq Matching (north={is_north})")
    return fig, ax


def plot_match_pointing(match: Match, show_pairs=True):
    fig, ax = plt.subplots(figsize=(10, 6))
    sa, da = match.src.as_array(), match.dst.as_array()
    plt.plot(np.rad2deg(da['xi']), np.rad2deg(da['eta']), '.')
    plt.plot(np.rad2deg(sa['xi']), np.rad2deg(sa['eta']), '.')
    if show_pairs:
        for r1, r2 in match.get_match_iter(include_unmatched=False):
            plt.plot(
                *np.rad2deg(np.array([[r1.xi, r2.xi], [r1.eta, r2.eta]])),
                color='red', alpha=.2
            )
    ax.set_xlabel("Xi (deg)")
    ax.set_ylabel("Eta (deg)")

    return fig, ax
